# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import itertools
import pytest
import random

from collections import defaultdict
from hypothesis import given

from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model.from_disk import DentryPerms
from swh.model.identifiers import (
    CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT
)

from swh.web.common import service
from swh.web.common.exc import BadInputExc, NotFoundExc
from swh.web.tests.data import random_sha1, random_content
from swh.web.tests.strategies import (
    content, unknown_content, contents, unknown_contents,
    contents_with_ctags, origin, new_origin, visit_dates, directory,
    unknown_directory, release, unknown_release, revision, unknown_revision,
    revisions, ancestor_revisions, non_ancestor_revisions, invalid_sha1,
    sha256, revision_with_submodules, empty_directory, new_revision,
    snapshot, unknown_snapshot
)
from swh.web.tests.conftest import ctags_json_missing, fossology_missing


@given(contents())
def test_lookup_multiple_hashes_all_present(contents):
    input_data = []
    expected_output = []
    for cnt in contents:
        input_data.append({'sha1': cnt['sha1']})
        expected_output.append({'sha1': cnt['sha1'],
                                'found': True})

    assert service.lookup_multiple_hashes(input_data) == expected_output


@given(contents(), unknown_contents())
def test_lookup_multiple_hashes_some_missing(contents, unknown_contents):
    input_contents = list(itertools.chain(contents, unknown_contents))
    random.shuffle(input_contents)

    input_data = []
    expected_output = []
    for cnt in input_contents:
        input_data.append({'sha1': cnt['sha1']})
        expected_output.append({'sha1': cnt['sha1'],
                                'found': cnt in contents})

    assert service.lookup_multiple_hashes(input_data) == expected_output


def test_lookup_hash_does_not_exist():
    unknown_content_ = random_content()

    actual_lookup = service.lookup_hash('sha1_git:%s' %
                                        unknown_content_['sha1_git'])

    assert actual_lookup == {'found': None, 'algo': 'sha1_git'}


@given(content())
def test_lookup_hash_exist(archive_data, content):
    actual_lookup = service.lookup_hash('sha1:%s' % content['sha1'])

    content_metadata = archive_data.content_get_metadata(content['sha1'])

    assert {'found': content_metadata, 'algo': 'sha1'} == actual_lookup


def test_search_hash_does_not_exist():
    unknown_content_ = random_content()

    actual_lookup = service.search_hash('sha1_git:%s' %
                                        unknown_content_['sha1_git'])

    assert {'found': False} == actual_lookup


@given(content())
def test_search_hash_exist(content):
    actual_lookup = service.search_hash('sha1:%s' % content['sha1'])

    assert {'found': True} == actual_lookup


@pytest.mark.skipif(ctags_json_missing,
                    reason="requires ctags with json output support")
@given(contents_with_ctags())
def test_lookup_content_ctags(indexer_data, contents_with_ctags):
    content_sha1 = random.choice(contents_with_ctags['sha1s'])
    indexer_data.content_add_ctags(content_sha1)
    actual_ctags = list(service.lookup_content_ctags('sha1:%s' % content_sha1))

    expected_data = list(indexer_data.content_get_ctags(content_sha1))
    for ctag in expected_data:
        ctag['id'] = content_sha1

    assert actual_ctags == expected_data


def test_lookup_content_ctags_no_hash():
    unknown_content_ = random_content()

    actual_ctags = list(service.lookup_content_ctags('sha1:%s' %
                                                     unknown_content_['sha1']))

    assert actual_ctags == []


@given(content())
def test_lookup_content_filetype(indexer_data, content):
    indexer_data.content_add_mimetype(content['sha1'])
    actual_filetype = service.lookup_content_filetype(content['sha1'])

    expected_filetype = indexer_data.content_get_mimetype(content['sha1'])
    assert actual_filetype == expected_filetype


@pytest.mark.skip  # Language indexer is disabled.
@given(content())
def test_lookup_content_language(indexer_data, content):
    indexer_data.content_add_language(content['sha1'])
    actual_language = service.lookup_content_language(content['sha1'])

    expected_language = indexer_data.content_get_language(content['sha1'])
    assert actual_language == expected_language


@given(contents_with_ctags())
def test_lookup_expression(indexer_data, contents_with_ctags):
    per_page = 10
    expected_ctags = []

    for content_sha1 in contents_with_ctags['sha1s']:
        if len(expected_ctags) == per_page:
            break
        indexer_data.content_add_ctags(content_sha1)
        for ctag in indexer_data.content_get_ctags(content_sha1):
            if len(expected_ctags) == per_page:
                break
            if ctag['name'] == contents_with_ctags['symbol_name']:
                del ctag['id']
                ctag['sha1'] = content_sha1
                expected_ctags.append(ctag)

    actual_ctags = list(
        service.lookup_expression(contents_with_ctags['symbol_name'],
                                  last_sha1=None, per_page=10))

    assert actual_ctags == expected_ctags


def test_lookup_expression_no_result():
    expected_ctags = []

    actual_ctags = list(service.lookup_expression('barfoo', last_sha1=None,
                                                  per_page=10))
    assert actual_ctags == expected_ctags


@pytest.mark.skipif(fossology_missing,
                    reason="requires fossology-nomossa installed")
@given(content())
def test_lookup_content_license(indexer_data, content):
    indexer_data.content_add_license(content['sha1'])
    actual_license = service.lookup_content_license(content['sha1'])

    expected_license = indexer_data.content_get_license(content['sha1'])
    assert actual_license == expected_license


def test_stat_counters(archive_data):
    actual_stats = service.stat_counters()
    assert actual_stats == archive_data.stat_counters()


@given(new_origin(), visit_dates())
def test_lookup_origin_visits(archive_data, new_origin, visit_dates):
    archive_data.origin_add_one(new_origin)
    for ts in visit_dates:
        archive_data.origin_visit_add(
            new_origin['url'], ts, type='git')

    actual_origin_visits = list(
        service.lookup_origin_visits(new_origin['url'], per_page=100))

    expected_visits = archive_data.origin_visit_get(new_origin['url'])
    for expected_visit in expected_visits:
        expected_visit['origin'] = new_origin['url']

    assert actual_origin_visits == expected_visits


@given(new_origin(), visit_dates())
def test_lookup_origin_visit(archive_data, new_origin, visit_dates):
    archive_data.origin_add_one(new_origin)
    visits = []
    for ts in visit_dates:
        visits.append(archive_data.origin_visit_add(
            new_origin['url'], ts, type='git'))

    visit = random.choice(visits)['visit']
    actual_origin_visit = service.lookup_origin_visit(
        new_origin['url'], visit)

    expected_visit = dict(archive_data.origin_visit_get_by(
        new_origin['url'], visit))

    assert actual_origin_visit == expected_visit


@given(new_origin())
def test_lookup_origin(archive_data, new_origin):
    archive_data.origin_add_one(new_origin)

    actual_origin = service.lookup_origin({'url': new_origin['url']})
    expected_origin = archive_data.origin_get(
        {'url': new_origin['url']})
    assert actual_origin == expected_origin


@given(invalid_sha1())
def test_lookup_release_ko_id_checksum_not_a_sha1(invalid_sha1):
    with pytest.raises(BadInputExc) as e:
        service.lookup_release(invalid_sha1)
    assert e.match('Invalid checksum')


@given(sha256())
def test_lookup_release_ko_id_checksum_too_long(sha256):
    with pytest.raises(BadInputExc) as e:
        service.lookup_release(sha256)
    assert e.match('Only sha1_git is supported.')


@given(directory())
def test_lookup_directory_with_path_not_found(directory):
    path = 'some/invalid/path/here'
    with pytest.raises(NotFoundExc) as e:
        service.lookup_directory_with_path(directory, path)
    assert e.match('Directory entry with path %s from %s not found' %
                   (path, directory))


@given(directory())
def test_lookup_directory_with_path_found(archive_data, directory):
    directory_content = archive_data.directory_ls(directory)
    directory_entry = random.choice(directory_content)
    path = directory_entry['name']
    actual_result = service.lookup_directory_with_path(directory, path)
    assert actual_result == directory_entry


@given(release())
def test_lookup_release(archive_data, release):
    actual_release = service.lookup_release(release)

    assert actual_release == archive_data.release_get(release)


@given(revision(), invalid_sha1(), sha256())
def test_lookup_revision_with_context_ko_not_a_sha1(revision,
                                                    invalid_sha1,
                                                    sha256):
    sha1_git_root = revision
    sha1_git = invalid_sha1

    with pytest.raises(BadInputExc) as e:
        service.lookup_revision_with_context(sha1_git_root, sha1_git)
    assert e.match('Invalid checksum query string')

    sha1_git = sha256

    with pytest.raises(BadInputExc) as e:
        service.lookup_revision_with_context(sha1_git_root, sha1_git)
    assert e.match('Only sha1_git is supported')


@given(revision(), unknown_revision())
def test_lookup_revision_with_context_ko_sha1_git_does_not_exist(
        revision, unknown_revision):
    sha1_git_root = revision
    sha1_git = unknown_revision

    with pytest.raises(NotFoundExc) as e:
        service.lookup_revision_with_context(sha1_git_root, sha1_git)
    assert e.match('Revision %s not found' % sha1_git)


@given(revision(), unknown_revision())
def test_lookup_revision_with_context_ko_root_sha1_git_does_not_exist(
        revision, unknown_revision):
    sha1_git_root = unknown_revision
    sha1_git = revision
    with pytest.raises(NotFoundExc) as e:
        service.lookup_revision_with_context(sha1_git_root, sha1_git)
    assert e.match('Revision root %s not found' % sha1_git_root)


@given(ancestor_revisions())
def test_lookup_revision_with_context(archive_data, ancestor_revisions):
    sha1_git = ancestor_revisions['sha1_git']
    root_sha1_git = ancestor_revisions['sha1_git_root']
    for sha1_git_root in (root_sha1_git,
                          {'id': hash_to_bytes(root_sha1_git)}):
        actual_revision = service.lookup_revision_with_context(sha1_git_root,
                                                               sha1_git)

        children = []
        for rev in archive_data.revision_log(root_sha1_git):
            for p_rev in rev['parents']:
                p_rev_hex = hash_to_hex(p_rev)
                if p_rev_hex == sha1_git:
                    children.append(rev['id'])

        expected_revision = archive_data.revision_get(sha1_git)
        expected_revision['children'] = children
        assert actual_revision == expected_revision


@given(non_ancestor_revisions())
def test_lookup_revision_with_context_ko(non_ancestor_revisions):
    sha1_git = non_ancestor_revisions['sha1_git']
    root_sha1_git = non_ancestor_revisions['sha1_git_root']

    with pytest.raises(NotFoundExc) as e:
        service.lookup_revision_with_context(root_sha1_git, sha1_git)
    assert e.match('Revision %s is not an ancestor of %s' %
                   (sha1_git, root_sha1_git))


def test_lookup_directory_with_revision_not_found():
    unknown_revision_ = random_sha1()

    with pytest.raises(NotFoundExc) as e:
        service.lookup_directory_with_revision(unknown_revision_)
    assert e.match('Revision %s not found' % unknown_revision_)


def test_lookup_directory_with_revision_unknown_content(archive_data):
    unknown_content_ = random_content()
    unknown_revision_ = random_sha1()
    unknown_directory_ = random_sha1()

    dir_path = 'README.md'
    # Create a revision that points to a directory
    # Which points to unknown content
    revision = {
        'author': {
            'name': b'abcd',
            'email': b'abcd@company.org',
            'fullname': b'abcd abcd'
        },
        'committer': {
            'email': b'aaaa@company.org',
            'fullname': b'aaaa aaa',
            'name': b'aaa'
        },
        'committer_date': {
            'negative_utc': False,
            'offset': 0,
            'timestamp': 1437511651
        },
        'date': {
            'negative_utc': False,
            'offset': 0,
            'timestamp': 1437511651
        },
        'message': b'bleh',
        'metadata': [],
        'parents': [],
        'synthetic': False,
        'type': 'git',
        'id': hash_to_bytes(unknown_revision_),
        'directory': hash_to_bytes(unknown_directory_)
    }
    # A directory that points to unknown content
    dir = {
        'id': hash_to_bytes(unknown_directory_),
        'entries': [{
            'name': bytes(dir_path.encode('utf-8')),
            'type': 'file',
            'target': hash_to_bytes(unknown_content_['sha1_git']),
            'perms': DentryPerms.content
        }]
    }
    # Add the directory and revision in mem
    archive_data.directory_add([dir])
    archive_data.revision_add([revision])
    with pytest.raises(NotFoundExc) as e:
        service.lookup_directory_with_revision(unknown_revision_, dir_path)
    assert e.match('Content not found for revision %s' % unknown_revision_)


@given(revision())
def test_lookup_directory_with_revision_ko_path_to_nowhere(revision):
    invalid_path = 'path/to/something/unknown'
    with pytest.raises(NotFoundExc) as e:
        service.lookup_directory_with_revision(revision, invalid_path)
    assert e.match('Directory or File')
    assert e.match(invalid_path)
    assert e.match('revision %s' % revision)
    assert e.match('not found')


@given(revision_with_submodules())
def test_lookup_directory_with_revision_submodules(archive_data,
                                                   revision_with_submodules):
    rev_sha1_git = revision_with_submodules['rev_sha1_git']
    rev_dir_path = revision_with_submodules['rev_dir_rev_path']

    actual_data = service.lookup_directory_with_revision(
        rev_sha1_git, rev_dir_path)

    revision = archive_data.revision_get(
        revision_with_submodules['rev_sha1_git'])
    directory = archive_data.directory_ls(revision['directory'])
    rev_entry = next(e for e in directory if e['name'] == rev_dir_path)

    expected_data = {
        'content': archive_data.revision_get(rev_entry['target']),
        'path': rev_dir_path,
        'revision': rev_sha1_git,
        'type': 'rev'
    }

    assert actual_data == expected_data


@given(revision())
def test_lookup_directory_with_revision_without_path(archive_data, revision):
    actual_directory_entries = service.lookup_directory_with_revision(revision)

    revision_data = archive_data.revision_get(revision)
    expected_directory_entries = archive_data.directory_ls(
        revision_data['directory'])

    assert actual_directory_entries['type'] == 'dir'
    assert actual_directory_entries['content'] == expected_directory_entries


@given(revision())
def test_lookup_directory_with_revision_with_path(archive_data, revision):
    rev_data = archive_data.revision_get(revision)
    dir_entries = [e for e in archive_data.directory_ls(rev_data['directory'])
                   if e['type'] in ('file', 'dir')]
    expected_dir_entry = random.choice(dir_entries)

    actual_dir_entry = service.lookup_directory_with_revision(
        revision, expected_dir_entry['name'])

    assert actual_dir_entry['type'] == expected_dir_entry['type']
    assert actual_dir_entry['revision'] == revision
    assert actual_dir_entry['path'] == expected_dir_entry['name']
    if actual_dir_entry['type'] == 'file':
        del actual_dir_entry['content']['checksums']['blake2s256']
        for key in ('checksums', 'status', 'length'):
            assert actual_dir_entry['content'][key] == expected_dir_entry[key]
    else:
        sub_dir_entries = archive_data.directory_ls(
            expected_dir_entry['target'])
        assert actual_dir_entry['content'] == sub_dir_entries


@given(revision())
def test_lookup_directory_with_revision_with_path_to_file_and_data(
        archive_data, revision):
    rev_data = archive_data.revision_get(revision)
    dir_entries = [e for e in archive_data.directory_ls(rev_data['directory'])
                   if e['type'] == 'file']
    expected_dir_entry = random.choice(dir_entries)
    expected_data = archive_data.content_get(
        expected_dir_entry['checksums']['sha1'])

    actual_dir_entry = service.lookup_directory_with_revision(
        revision, expected_dir_entry['name'], with_data=True)

    assert actual_dir_entry['type'] == expected_dir_entry['type']
    assert actual_dir_entry['revision'] == revision
    assert actual_dir_entry['path'] == expected_dir_entry['name']
    del actual_dir_entry['content']['checksums']['blake2s256']
    for key in ('checksums', 'status', 'length'):
        assert actual_dir_entry['content'][key] == expected_dir_entry[key]
    assert actual_dir_entry['content']['data'] == expected_data['data']


@given(revision())
def test_lookup_revision(archive_data, revision):
    actual_revision = service.lookup_revision(revision)
    assert actual_revision == archive_data.revision_get(revision)


@given(new_revision())
def test_lookup_revision_invalid_msg(archive_data, new_revision):
    new_revision['message'] = b'elegant fix for bug \xff'
    archive_data.revision_add([new_revision])

    revision = service.lookup_revision(hash_to_hex(new_revision['id']))
    assert revision['message'] is None
    assert revision['message_decoding_failed'] is True


@given(new_revision())
def test_lookup_revision_msg_ok(archive_data, new_revision):
    archive_data.revision_add([new_revision])

    revision_message = service.lookup_revision_message(
        hash_to_hex(new_revision['id']))

    assert revision_message == {'message': new_revision['message']}


def test_lookup_revision_msg_no_rev():
    unknown_revision_ = random_sha1()

    with pytest.raises(NotFoundExc) as e:
        service.lookup_revision_message(unknown_revision_)

    assert e.match('Revision with sha1_git %s not found.' % unknown_revision_)


@given(revisions())
def test_lookup_revision_multiple(archive_data, revisions):
    actual_revisions = list(service.lookup_revision_multiple(revisions))

    expected_revisions = []
    for rev in revisions:
        expected_revisions.append(archive_data.revision_get(rev))

    assert actual_revisions == expected_revisions


def test_lookup_revision_multiple_none_found():
    unknown_revisions_ = [random_sha1(), random_sha1(), random_sha1()]

    actual_revisions = list(
        service.lookup_revision_multiple(unknown_revisions_))

    assert actual_revisions == [None] * len(unknown_revisions_)


@given(revision())
def test_lookup_revision_log(archive_data, revision):
    actual_revision_log = list(service.lookup_revision_log(revision, limit=25))
    expected_revision_log = archive_data.revision_log(revision, limit=25)

    assert actual_revision_log == expected_revision_log


def _get_origin_branches(archive_data, origin):
    origin_visit = archive_data.origin_visit_get(origin['url'])[-1]
    snapshot = archive_data.snapshot_get(origin_visit['snapshot'])
    branches = {k: v for (k, v) in snapshot['branches'].items()
                if v['target_type'] == 'revision'}
    return branches


@given(origin())
def test_lookup_revision_log_by(archive_data, origin):
    branches = _get_origin_branches(archive_data, origin)
    branch_name = random.choice(list(branches.keys()))

    actual_log = list(
        service.lookup_revision_log_by(origin['url'], branch_name,
                                       None, limit=25))

    expected_log = archive_data.revision_log(
        branches[branch_name]['target'], limit=25)

    assert actual_log == expected_log


@given(origin())
def test_lookup_revision_log_by_notfound(origin):
    with pytest.raises(NotFoundExc):
        service.lookup_revision_log_by(
            origin['url'], 'unknown_branch_name', None, limit=100)


def test_lookup_content_raw_not_found():
    unknown_content_ = random_content()

    with pytest.raises(NotFoundExc) as e:
        service.lookup_content_raw('sha1:' + unknown_content_['sha1'])

    assert e.match('Content with %s checksum equals to %s not found!' %
                   ('sha1', unknown_content_['sha1']))


@given(content())
def test_lookup_content_raw(archive_data, content):
    actual_content = service.lookup_content_raw(
        'sha256:%s' % content['sha256'])

    expected_content = archive_data.content_get(content['sha1'])

    assert actual_content == expected_content


def test_lookup_content_not_found():
    unknown_content_ = random_content()

    with pytest.raises(NotFoundExc) as e:
        service.lookup_content('sha1:%s' % unknown_content_['sha1'])

    assert e.match('Content with %s checksum equals to %s not found!' %
                   ('sha1', unknown_content_['sha1']))


@given(content())
def test_lookup_content_with_sha1(archive_data, content):
    actual_content = service.lookup_content('sha1:%s' % content['sha1'])

    expected_content = archive_data.content_get_metadata(content['sha1'])

    assert actual_content == expected_content


@given(content())
def test_lookup_content_with_sha256(archive_data, content):
    actual_content = service.lookup_content('sha256:%s' % content['sha256'])

    expected_content = archive_data.content_get_metadata(content['sha1'])

    assert actual_content == expected_content


def test_lookup_directory_bad_checksum():
    with pytest.raises(BadInputExc):
        service.lookup_directory('directory_id')


def test_lookup_directory_not_found():
    unknown_directory_ = random_sha1()

    with pytest.raises(NotFoundExc) as e:
        service.lookup_directory(unknown_directory_)

    assert e.match('Directory with sha1_git %s not found' % unknown_directory_)


@given(directory())
def test_lookup_directory(archive_data, directory):
    actual_directory_ls = list(service.lookup_directory(directory))

    expected_directory_ls = archive_data.directory_ls(directory)

    assert actual_directory_ls == expected_directory_ls


@given(empty_directory())
def test_lookup_directory_empty(empty_directory):
    actual_directory_ls = list(service.lookup_directory(empty_directory))

    assert actual_directory_ls == []


@given(origin())
def test_lookup_revision_by_nothing_found(origin):
    with pytest.raises(NotFoundExc):
        service.lookup_revision_by(origin['url'], 'invalid-branch-name')


@given(origin())
def test_lookup_revision_by(archive_data, origin):
    branches = _get_origin_branches(archive_data, origin)
    branch_name = random.choice(list(branches.keys()))

    actual_revision = service.lookup_revision_by(origin['url'], branch_name)

    expected_revision = archive_data.revision_get(
        branches[branch_name]['target'])

    assert actual_revision == expected_revision


@given(origin(), revision())
def test_lookup_revision_with_context_by_ko(origin, revision):
    with pytest.raises(NotFoundExc):
        service.lookup_revision_with_context_by(origin['url'],
                                                'invalid-branch-name',
                                                None, revision)


@given(origin())
def test_lookup_revision_with_context_by(archive_data, origin):
    branches = _get_origin_branches(archive_data, origin)
    branch_name = random.choice(list(branches.keys()))

    root_rev = branches[branch_name]['target']
    root_rev_log = archive_data.revision_log(root_rev)

    children = defaultdict(list)

    for rev in root_rev_log:
        for rev_p in rev['parents']:
            children[rev_p].append(rev['id'])

    rev = root_rev_log[-1]['id']

    actual_root_rev, actual_rev = service.lookup_revision_with_context_by(
        origin['url'], branch_name, None, rev)

    expected_root_rev = archive_data.revision_get(root_rev)
    expected_rev = archive_data.revision_get(rev)
    expected_rev['children'] = children[rev]

    assert actual_root_rev == expected_root_rev
    assert actual_rev == expected_rev


def test_lookup_revision_through_ko_not_implemented():
    with pytest.raises(NotImplementedError):
        service.lookup_revision_through({'something-unknown': 10})


@given(origin())
def test_lookup_revision_through_with_context_by(archive_data, origin):
    branches = _get_origin_branches(archive_data, origin)
    branch_name = random.choice(list(branches.keys()))

    root_rev = branches[branch_name]['target']
    root_rev_log = archive_data.revision_log(root_rev)
    rev = root_rev_log[-1]['id']

    assert service.lookup_revision_through({
        'origin_url': origin['url'],
        'branch_name': branch_name,
        'ts': None,
        'sha1_git': rev
    }) == service.lookup_revision_with_context_by(origin['url'], branch_name,
                                                  None, rev)


@given(origin())
def test_lookup_revision_through_with_revision_by(archive_data, origin):
    branches = _get_origin_branches(archive_data, origin)
    branch_name = random.choice(list(branches.keys()))

    assert service.lookup_revision_through({
        'origin_url': origin['url'],
        'branch_name': branch_name,
        'ts': None,
    }) == service.lookup_revision_by(origin['url'], branch_name, None)


@given(ancestor_revisions())
def test_lookup_revision_through_with_context(ancestor_revisions):
    sha1_git = ancestor_revisions['sha1_git']
    sha1_git_root = ancestor_revisions['sha1_git_root']

    assert service.lookup_revision_through({
        'sha1_git_root': sha1_git_root,
        'sha1_git': sha1_git,
    }) == service.lookup_revision_with_context(sha1_git_root, sha1_git)


@given(revision())
def test_lookup_revision_through_with_revision(revision):
    assert service.lookup_revision_through({
        'sha1_git': revision
    }) == service.lookup_revision(revision)


@given(revision())
def test_lookup_directory_through_revision_ko_not_found(revision):
    with pytest.raises(NotFoundExc):
        service.lookup_directory_through_revision(
            {'sha1_git': revision}, 'some/invalid/path')


@given(revision())
def test_lookup_directory_through_revision_ok(archive_data, revision):
    rev_data = archive_data.revision_get(revision)
    dir_entries = [e for e in archive_data.directory_ls(rev_data['directory'])
                   if e['type'] == 'file']
    dir_entry = random.choice(dir_entries)

    assert service.lookup_directory_through_revision(
        {'sha1_git': revision}, dir_entry['name']
    ) == (revision,
          service.lookup_directory_with_revision(revision, dir_entry['name']))


@given(revision())
def test_lookup_directory_through_revision_ok_with_data(
        archive_data, revision):
    rev_data = archive_data.revision_get(revision)
    dir_entries = [e for e in archive_data.directory_ls(rev_data['directory'])
                   if e['type'] == 'file']
    dir_entry = random.choice(dir_entries)

    assert service.lookup_directory_through_revision(
        {'sha1_git': revision}, dir_entry['name'], with_data=True
    ) == (revision,
          service.lookup_directory_with_revision(revision, dir_entry['name'],
                                                 with_data=True))


@given(content(), directory(), release(), revision(), snapshot())
def test_lookup_known_objects(archive_data, content, directory, release,
                              revision, snapshot):
    expected = archive_data.content_find(content)
    assert service.lookup_object(CONTENT, content['sha1_git']) == expected

    expected = archive_data.directory_get(directory)
    assert service.lookup_object(DIRECTORY, directory) == expected

    expected = archive_data.release_get(release)
    assert service.lookup_object(RELEASE, release) == expected

    expected = archive_data.revision_get(revision)
    assert service.lookup_object(REVISION, revision) == expected

    expected = archive_data.snapshot_get(snapshot)
    assert service.lookup_object(SNAPSHOT, snapshot) == expected


@given(unknown_content(), unknown_directory(), unknown_release(),
       unknown_revision(), unknown_snapshot())
def test_lookup_unknown_objects(unknown_content, unknown_directory,
                                unknown_release, unknown_revision,
                                unknown_snapshot):
    with pytest.raises(NotFoundExc) as e:
        service.lookup_object(CONTENT, unknown_content['sha1_git'])
    assert e.match(r'Content.*not found')

    with pytest.raises(NotFoundExc) as e:
        service.lookup_object(DIRECTORY, unknown_directory)
    assert e.match(r'Directory.*not found')

    with pytest.raises(NotFoundExc) as e:
        service.lookup_object(RELEASE, unknown_release)
    assert e.match(r'Release.*not found')

    with pytest.raises(NotFoundExc) as e:
        service.lookup_object(REVISION, unknown_revision)
    assert e.match(r'Revision.*not found')

    with pytest.raises(NotFoundExc) as e:
        service.lookup_object(SNAPSHOT, unknown_snapshot)
    assert e.match(r'Snapshot.*not found')


@given(invalid_sha1())
def test_lookup_invalid_objects(invalid_sha1):

    with pytest.raises(BadInputExc) as e:
        service.lookup_object('foo', invalid_sha1)
    assert e.match('Invalid swh object type')

    with pytest.raises(BadInputExc) as e:
        service.lookup_object(CONTENT, invalid_sha1)
    assert e.match('Invalid hash')

    with pytest.raises(BadInputExc) as e:
        service.lookup_object(DIRECTORY, invalid_sha1)
    assert e.match('Invalid checksum')

    with pytest.raises(BadInputExc) as e:
        service.lookup_object(RELEASE, invalid_sha1)
    assert e.match('Invalid checksum')

    with pytest.raises(BadInputExc) as e:
        service.lookup_object(REVISION, invalid_sha1)
    assert e.match('Invalid checksum')

    with pytest.raises(BadInputExc) as e:
        service.lookup_object(SNAPSHOT, invalid_sha1)
    assert e.match('Invalid checksum')



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.shortcuts import render, redirect

from swh.web.common import service
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import (
    reverse, format_utc_iso_date, parse_timestamp
)
from swh.web.common.exc import handle_view_exception
from swh.web.browse.utils import get_snapshot_context
from swh.web.browse.browseurls import browse_route

from .utils.snapshot_context import (
    browse_snapshot_directory, browse_snapshot_content,
    browse_snapshot_log, browse_snapshot_branches,
    browse_snapshot_releases
)


@browse_route(r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/directory/',
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)'
              '/directory/(?P<path>.+)/',
              r'origin/(?P<origin_url>.+)/directory/',
              r'origin/(?P<origin_url>.+)/directory/(?P<path>.+)/',
              view_name='browse-origin-directory')
def origin_directory_browse(request, origin_url,
                            timestamp=None, path=None):
    """Django view for browsing the content of a directory associated
    to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/(origin_url)/directory/[(path)/]`
        * :http:get:`/browse/origin/(origin_url)/visit/(timestamp)/directory/[(path)/]`
    """ # noqa
    return browse_snapshot_directory(request, origin_url=origin_url,
                                     timestamp=timestamp, path=path)


@browse_route(r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)'
              '/content/(?P<path>.+)/',
              r'origin/(?P<origin_url>.+)/content/(?P<path>.+)/',
              view_name='browse-origin-content')
def origin_content_browse(request, origin_url, path=None,
                          timestamp=None):
    """Django view that produces an HTML display of a content
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/(origin_url)/content/(path)/`
        * :http:get:`/browse/origin/(origin_url)/visit/(timestamp)/content/(path)/`

    """ # noqa
    language = request.GET.get('language', None)
    return browse_snapshot_content(request,
                                   origin_url=origin_url, timestamp=timestamp,
                                   path=path, selected_language=language)


PER_PAGE = 20


@browse_route(r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/log/',
              r'origin/(?P<origin_url>.+)/log/',
              view_name='browse-origin-log')
def origin_log_browse(request, origin_url, timestamp=None):
    """Django view that produces an HTML display of revisions history (aka
    the commit log) associated to a software origin.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/(origin_url)/log/`
        * :http:get:`/browse/origin/(origin_url)/visit/(timestamp)/log/`
    """ # noqa
    return browse_snapshot_log(request,
                               origin_url=origin_url, timestamp=timestamp)


@browse_route(r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/branches/',
              r'origin/(?P<origin_url>.+)/branches/',
              view_name='browse-origin-branches')
def origin_branches_browse(request, origin_url, timestamp=None):
    """Django view that produces an HTML display of the list of branches
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/(origin_url)/branches/`
        * :http:get:`/browse/origin/(origin_url)/visit/(timestamp)/branches/`

    """ # noqa
    return browse_snapshot_branches(request,
                                    origin_url=origin_url, timestamp=timestamp)


@browse_route(r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/releases/',
              r'origin/(?P<origin_url>.+)/releases/',
              view_name='browse-origin-releases')
def origin_releases_browse(request, origin_url, timestamp=None):
    """Django view that produces an HTML display of the list of releases
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/(origin_url)/releases/`
        * :http:get:`/browse/origin/(origin_url)/visit/(timestamp)/releases/`

    """ # noqa
    return browse_snapshot_releases(request,
                                    origin_url=origin_url, timestamp=timestamp)


@browse_route(r'origin/(?P<origin_url>.+)/visits/',
              view_name='browse-origin-visits')
def origin_visits_browse(request, origin_url):
    """Django view that produces an HTML display of visits reporting
    for a swh origin identified by its id or its url.

    The url that points to it is
    :http:get:`/browse/origin/(origin_url)/visits/`.
    """
    try:
        origin_info = service.lookup_origin({'url': origin_url})
        origin_visits = get_origin_visits(origin_info)
        snapshot_context = get_snapshot_context(origin_url=origin_url)
    except Exception as exc:
        return handle_view_exception(request, exc)

    for i, visit in enumerate(origin_visits):
        url_date = format_utc_iso_date(visit['date'], '%Y-%m-%dT%H:%M:%SZ')
        visit['fmt_date'] = format_utc_iso_date(visit['date'])
        query_params = {}
        if i < len(origin_visits) - 1:
            if visit['date'] == origin_visits[i+1]['date']:
                query_params = {'visit_id': visit['visit']}
        if i > 0:
            if visit['date'] == origin_visits[i-1]['date']:
                query_params = {'visit_id': visit['visit']}

        snapshot = visit['snapshot'] if visit['snapshot'] else ''

        visit['browse_url'] = reverse('browse-origin-directory',
                                      url_args={'origin_url': origin_url,
                                                'timestamp': url_date},
                                      query_params=query_params)
        if not snapshot:
            visit['snapshot'] = ''
        visit['date'] = parse_timestamp(visit['date']).timestamp()

    heading = 'Origin visits - %s' % origin_url

    return render(request, 'browse/origin-visits.html',
                  {'heading': heading,
                   'swh_object_name': 'Visits',
                   'swh_object_metadata': origin_info,
                   'origin_visits': origin_visits,
                   'origin_info': origin_info,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': False})


@browse_route(r'origin/(?P<origin_url>.+)/',
              view_name='browse-origin')
def origin_browse(request, origin_url):
    """Django view that redirects to the display of the latest archived
    snapshot for a given software origin.
    """
    last_snapshot_url = reverse('browse-origin-directory',
                                url_args={'origin_url': origin_url})
    return redirect(last_snapshot_url)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given

from swh.model.identifiers import (
    CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT
)

from swh.web.common.utils import reverse
from swh.web.tests.strategies import (
    content, directory, origin, release, revision, snapshot,
    unknown_content, unknown_directory, unknown_release,
    unknown_revision, unknown_snapshot
)


@given(origin(), content(), directory(), release(), revision(), snapshot())
def test_swh_id_resolve_success(api_client, origin, content, directory,
                                release, revision, snapshot):

    for obj_type_short, obj_type, obj_id in (
            ('cnt', CONTENT, content['sha1_git']),
            ('dir', DIRECTORY, directory),
            ('rel', RELEASE, release),
            ('rev', REVISION, revision),
            ('snp', SNAPSHOT, snapshot)):

        swh_id = 'swh:1:%s:%s;origin=%s' % (obj_type_short, obj_id,
                                            origin['url'])
        url = reverse('api-1-resolve-swh-pid', url_args={'swh_id': swh_id})

        resp = api_client.get(url)

        if obj_type == CONTENT:
            url_args = {'query_string': 'sha1_git:%s' % obj_id}
        elif obj_type == SNAPSHOT:
            url_args = {'snapshot_id': obj_id}
        else:
            url_args = {'sha1_git': obj_id}

        browse_rev_url = reverse('browse-%s' % obj_type,
                                 url_args=url_args,
                                 query_params={'origin': origin['url']},
                                 request=resp.wsgi_request)

        expected_result = {
            'browse_url': browse_rev_url,
            'metadata': {'origin': origin['url']},
            'namespace': 'swh',
            'object_id': obj_id,
            'object_type': obj_type,
            'scheme_version': 1
        }

        assert resp.status_code == 200, resp.data
        assert resp.data == expected_result


def test_swh_id_resolve_invalid(api_client):
    rev_id_invalid = '96db9023b8_foo_50d6c108e9a3'
    swh_id = 'swh:1:rev:%s' % rev_id_invalid
    url = reverse('api-1-resolve-swh-pid', url_args={'swh_id': swh_id})

    resp = api_client.get(url)

    assert resp.status_code == 400, resp.data


@given(unknown_content(), unknown_directory(), unknown_release(),
       unknown_revision(), unknown_snapshot())
def test_swh_id_resolve_not_found(api_client, unknown_content,
                                  unknown_directory, unknown_release,
                                  unknown_revision, unknown_snapshot):

    for obj_type_short, obj_id in (('cnt', unknown_content['sha1_git']),
                                   ('dir', unknown_directory),
                                   ('rel', unknown_release),
                                   ('rev', unknown_revision),
                                   ('snp', unknown_snapshot)):

        swh_id = 'swh:1:%s:%s' % (obj_type_short, obj_id)

        url = reverse('api-1-resolve-swh-pid', url_args={'swh_id': swh_id})

        resp = api_client.get(url)

        assert resp.status_code == 404, resp.data


def test_swh_origin_id_not_resolvable(api_client):
    ori_pid = 'swh:1:ori:8068d0075010b590762c6cb5682ed53cb3c13deb'
    url = reverse('api-1-resolve-swh-pid', url_args={'swh_id': ori_pid})
    resp = api_client.get(url)
    assert resp.status_code == 400, resp.data



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Django tests settings for swh-web.
"""

import sys

from swh.web.config import get_config

scope1_limiter_rate = 3
scope1_limiter_rate_post = 1
scope2_limiter_rate = 5
scope2_limiter_rate_post = 2
scope3_limiter_rate = 1
scope3_limiter_rate_post = 1
save_origin_rate_post = 10

swh_web_config = get_config()

swh_web_config.update({
    'debug': False,
    'secret_key': 'test',
    'history_counters_url': '',
    'throttling': {
        'cache_uri': None,
        'scopes': {
            'swh_api': {
                'limiter_rate': {
                    'default': '60/min'
                },
                'exempted_networks': ['127.0.0.0/8']
            },
            'swh_api_origin_search': {
                'limiter_rate': {
                    'default': '100/min'
                },
                'exempted_networks': ['127.0.0.0/8']
            },
            'swh_api_origin_visit_latest': {
                'limiter_rate': {
                    'default': '6000/min'
                },
                'exempted_networks': ['127.0.0.0/8']
            },
            'swh_vault_cooking': {
                'limiter_rate': {
                    'default': '120/h',
                    'GET': '60/m'
                },
                'exempted_networks': ['127.0.0.0/8']
            },
            'swh_save_origin': {
                'limiter_rate': {
                    'default': '120/h',
                    'POST': '%s/h' % save_origin_rate_post,
                }
            },
            'scope1': {
                'limiter_rate': {
                    'default': '%s/min' % scope1_limiter_rate,
                    'POST': '%s/min' % scope1_limiter_rate_post,
                }
            },
            'scope2': {
                'limiter_rate': {
                    'default': '%s/min' % scope2_limiter_rate,
                    'POST': '%s/min' % scope2_limiter_rate_post
                }
            },
            'scope3': {
                'limiter_rate': {
                    'default': '%s/min' % scope3_limiter_rate,
                    'POST': '%s/min' % scope3_limiter_rate_post
                },
                'exempted_networks': ['127.0.0.0/8']
            }
        }
    }
})


from .common import * # noqa
from .common import ALLOWED_HOSTS, LOGGING # noqa

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': swh_web_config['test_db'],
    }
}

# when not running unit tests, make the webapp fetch data from memory storages
if 'pytest' not in sys.argv[0]:
    swh_web_config.update({
        'debug': True,
        'e2e_tests_mode': True
    })
    from swh.web.tests.data import get_tests_data, override_storages # noqa
    test_data = get_tests_data()
    override_storages(test_data['storage'], test_data['idx_storage'],
                      test_data['search'])
else:
    ALLOWED_HOSTS += ['testserver']

    # Silent DEBUG output when running unit tests
    LOGGING['handlers']['console']['level'] = 'INFO'  # type: ignore



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given

from swh.web.browse import utils
from swh.web.common.utils import reverse, format_utc_iso_date
from swh.web.tests.strategies import origin_with_multiple_visits


def test_get_mimetype_and_encoding_for_content():
    text = b'Hello world!'
    assert (utils.get_mimetype_and_encoding_for_content(text) ==
            ('text/plain', 'us-ascii'))


@given(origin_with_multiple_visits())
def test_get_origin_visit_snapshot_simple(archive_data, origin):
    visits = archive_data.origin_visit_get(origin['url'])

    for visit in visits:

        snapshot = archive_data.snapshot_get(visit['snapshot'])
        branches = []
        releases = []

        def _process_branch_data(branch, branch_data):
            if branch_data['target_type'] == 'revision':
                rev_data = archive_data.revision_get(branch_data['target'])
                branches.append({
                    'name': branch,
                    'revision': branch_data['target'],
                    'directory': rev_data['directory'],
                    'date': format_utc_iso_date(rev_data['date']),
                    'message': rev_data['message']
                })
            elif branch_data['target_type'] == 'release':
                rel_data = archive_data.release_get(branch_data['target'])
                rev_data = archive_data.revision_get(rel_data['target'])
                releases.append({
                    'name': rel_data['name'],
                    'branch_name': branch,
                    'date': format_utc_iso_date(rel_data['date']),
                    'id': rel_data['id'],
                    'message': rel_data['message'],
                    'target_type': rel_data['target_type'],
                    'target': rel_data['target'],
                    'directory': rev_data['directory']
                })

        for branch in sorted(snapshot['branches'].keys()):
            branch_data = snapshot['branches'][branch]
            if branch_data['target_type'] == 'alias':
                target_data = snapshot['branches'][branch_data['target']]
                _process_branch_data(branch, target_data)
            else:
                _process_branch_data(branch, branch_data)

        assert branches and releases, 'Incomplete test data.'

        origin_visit_branches = utils.get_origin_visit_snapshot(
            origin, visit_id=visit['visit'])

        assert origin_visit_branches == (branches, releases)


def test_gen_link():
    assert (utils.gen_link('https://www.softwareheritage.org/', 'swh') ==
            '<a href="https://www.softwareheritage.org/">swh</a>')


def test_gen_revision_link():
    revision_id = '28a0bc4120d38a394499382ba21d6965a67a3703'
    revision_url = reverse('browse-revision',
                           url_args={'sha1_git': revision_id})

    assert (utils.gen_revision_link(revision_id, link_text=None,
                                    link_attrs=None) ==
            '<a href="%s">%s</a>' % (revision_url, revision_id))
    assert (utils.gen_revision_link(revision_id, shorten_id=True,
                                    link_attrs=None) ==
            '<a href="%s">%s</a>' % (revision_url, revision_id[:7]))


def test_gen_person_mail_link():
    person_full = {
        'name': 'John Doe',
        'email': 'john.doe@swh.org',
        'fullname': 'John Doe <john.doe@swh.org>'
    }

    assert (utils.gen_person_mail_link(person_full) ==
            '<a href="mailto:%s">%s</a>' % (person_full['email'],
                                            person_full['name']))

    link_text = 'Mail'
    assert (utils.gen_person_mail_link(person_full, link_text=link_text) ==
            '<a href="mailto:%s">%s</a>' % (person_full['email'],
                                            link_text))

    person_partial_email = {
        'name': None,
        'email': None,
        'fullname': 'john.doe@swh.org'
    }

    assert (utils.gen_person_mail_link(person_partial_email) ==
            '<a href="mailto:%s">%s</a>' % (person_partial_email['fullname'],
                                            person_partial_email['fullname']))

    person_partial = {
        'name': None,
        'email': None,
        'fullname': 'John Doe <john.doe@swh.org>'
    }

    assert (utils.gen_person_mail_link(person_partial) ==
            person_partial['fullname'])

    person_none = {
        'name': None,
        'email': None,
        'fullname': None
    }

    assert utils.gen_person_mail_link(person_none) == 'None'



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from datetime import datetime

from django.test import Client

from swh.web.common.origin_save import (
    SAVE_REQUEST_ACCEPTED, SAVE_TASK_NOT_YET_SCHEDULED
)
from swh.web.common.utils import reverse
from swh.web.settings.tests import save_origin_rate_post


visit_type = 'git'
origin = {
    'url': 'https://github.com/python/cpython'
}


@pytest.fixture
def client():
    return Client(enforce_csrf_checks=True)


def test_save_request_form_csrf_token(client, mocker):
    mock_create_save_origin_request = mocker.patch(
        'swh.web.misc.origin_save.create_save_origin_request')
    _mock_create_save_origin_request(mock_create_save_origin_request)

    url = reverse('origin-save-request',
                  url_args={'visit_type': visit_type,
                            'origin_url': origin['url']})

    resp = client.post(url)
    assert resp.status_code == 403

    data = _get_csrf_token(client, reverse('origin-save'))
    resp = client.post(url, data=data)
    assert resp.status_code == 200


def test_save_request_form_rate_limit(client, mocker):
    mock_create_save_origin_request = mocker.patch(
        'swh.web.misc.origin_save.create_save_origin_request')
    _mock_create_save_origin_request(mock_create_save_origin_request)

    url = reverse('origin-save-request',
                  url_args={'visit_type': visit_type,
                            'origin_url':  origin['url']})

    data = _get_csrf_token(client, reverse('origin-save'))
    for _ in range(save_origin_rate_post):
        resp = client.post(url, data=data)
        assert resp.status_code == 200

    resp = client.post(url, data=data)
    assert resp.status_code == 429


def test_save_request_form_server_error(client, mocker):
    mock_create_save_origin_request = mocker.patch(
        'swh.web.misc.origin_save.create_save_origin_request')
    mock_create_save_origin_request.side_effect = Exception('Server error')

    url = reverse('origin-save-request',
                  url_args={'visit_type': visit_type,
                            'origin_url': origin['url']})

    data = _get_csrf_token(client, reverse('origin-save'))

    resp = client.post(url, data=data)
    assert resp.status_code == 500


def test_old_save_url_redirection(client):
    url = reverse('browse-origin-save')
    resp = client.get(url)
    assert resp.status_code == 302
    redirect_url = reverse('origin-save')
    assert resp['location'] == redirect_url


def _get_csrf_token(client, url):
    resp = client.get(url)
    return {
        'csrfmiddlewaretoken': resp.cookies['csrftoken'].value
    }


def _mock_create_save_origin_request(mock):
    expected_data = {
        'visit_type': visit_type,
        'origin_url': origin['url'],
        'save_request_date': datetime.now().isoformat(),
        'save_request_status': SAVE_REQUEST_ACCEPTED,
        'save_task_status': SAVE_TASK_NOT_YET_SCHEDULED,
        'visit_date': None
    }
    mock.return_value = expected_data



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect
import json
import os
import shutil
import sys

from subprocess import run, PIPE

import pytest

from django.core.cache import cache
from hypothesis import settings, HealthCheck
from rest_framework.test import APIClient, APIRequestFactory

from swh.model.hashutil import ALGORITHMS, hash_to_bytes
from swh.web.common import converters
from swh.web.tests.data import get_tests_data, override_storages

# Used to skip some tests
ctags_json_missing = (
    shutil.which('ctags') is None or
    b'+json' not in run(['ctags', '--version'], stdout=PIPE).stdout
)

fossology_missing = shutil.which('nomossa') is None

# Register some hypothesis profiles
settings.register_profile('default', settings())

settings.register_profile(
    'swh-web',
    settings(deadline=None,
             suppress_health_check=[HealthCheck.too_slow,
                                    HealthCheck.filter_too_much]))

settings.register_profile(
    'swh-web-fast',
    settings(deadline=None, max_examples=1,
             suppress_health_check=[HealthCheck.too_slow,
                                    HealthCheck.filter_too_much]))


def pytest_configure(config):
    # Small hack in order to be able to run the unit tests
    # without static assets generated by webpack.
    # Those assets are not really needed for the Python tests
    # but the django templates will fail to load due to missing
    # generated file webpack-stats.json describing the js and css
    # files to include.
    # So generate a dummy webpack-stats.json file to overcome
    # that issue.
    test_dir = os.path.dirname(__file__)
    # location of the static folder when running tests through tox
    static_dir = os.path.join(sys.prefix, 'share/swh/web/static')

    if not os.path.exists(static_dir):
        # location of the static folder when running tests locally with pytest
        static_dir = os.path.join(test_dir, '../../../static')
    webpack_stats = os.path.join(static_dir, 'webpack-stats.json')
    if os.path.exists(webpack_stats):
        return
    bundles_dir = os.path.join(test_dir, '../assets/src/bundles')
    _, dirs, _ = next(os.walk(bundles_dir))
    mock_webpack_stats = {
        'status': 'done',
        'publicPath': '/static',
        'chunks': {}
    }
    for bundle in dirs:
        asset = 'js/%s.js' % bundle
        mock_webpack_stats['chunks'][bundle] = [{
            'name': asset,
            'publicPath': '/static/%s' % asset,
            'path': os.path.join(static_dir, asset)
        }]

    with open(webpack_stats, 'w') as outfile:
        json.dump(mock_webpack_stats, outfile)


# Clear Django cache before each test
@pytest.fixture(autouse=True)
def django_cache_cleared():
    cache.clear()


# Alias rf fixture from pytest-django
@pytest.fixture
def request_factory(rf):
    return rf

# Fixture to get test client from Django REST Framework
@pytest.fixture(scope='module')
def api_client():
    return APIClient()


# Fixture to get API request factory from Django REST Framework
@pytest.fixture(scope='module')
def api_request_factory():
    return APIRequestFactory()


# Initialize tests data
@pytest.fixture(autouse=True)
def tests_data():
    data = get_tests_data(reset=True)
    # Update swh-web configuration to use the in-memory storages
    # instantiated in the tests.data module
    override_storages(data['storage'], data['idx_storage'], data['search'])
    return data


# Fixture to manipulate data from a sample archive used in the tests
@pytest.fixture
def archive_data(tests_data):
    return _ArchiveData(tests_data)


# Fixture to manipulate indexer data from a sample archive used in the tests
@pytest.fixture
def indexer_data(tests_data):
    return _IndexerData(tests_data)


# Custom data directory for requests_mock
@pytest.fixture
def datadir():
    return os.path.join(os.path.abspath(os.path.dirname(__file__)),
                        'resources')


class _ArchiveData:
    """
    Helper class to manage data from a sample test archive.

    It is initialized with a reference to an in-memory storage
    containing raw tests data.

    It is basically a proxy to Storage interface but it overrides some methods
    to retrieve those tests data in a json serializable format in order to ease
    tests implementation.
    """

    def __init__(self, tests_data):
        self.storage = tests_data['storage']

        def _call_storage_method(method):
            def call_storage_method(*args, **kwargs):
                return method(*args, **kwargs)

            return call_storage_method

        # Forward calls to non overridden Storage methods to wrapped
        # storage instance
        for method_name, method in inspect.getmembers(
                self.storage, predicate=inspect.ismethod):
            if (not hasattr(self, method_name) and
                    not method_name.startswith('_')):
                setattr(self, method_name, _call_storage_method(method))

    def content_find(self, content):
        cnt_ids_bytes = {algo_hash: hash_to_bytes(content[algo_hash])
                         for algo_hash in ALGORITHMS
                         if content.get(algo_hash)}
        cnt = self.storage.content_find(cnt_ids_bytes)
        return converters.from_content(cnt[0]) if cnt else cnt

    def content_get_metadata(self, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        metadata = self.storage.content_get_metadata([cnt_id_bytes])
        contents = metadata[cnt_id_bytes]
        content = None if not contents else contents[0]
        return converters.from_swh(content,
                                   hashess={'sha1', 'sha1_git', 'sha256',
                                            'blake2s256'})

    def content_get(self, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        cnt = next(self.storage.content_get([cnt_id_bytes]))
        return converters.from_content(cnt)

    def directory_get(self, dir_id):
        return {
            'id': dir_id,
            'content': self.directory_ls(dir_id)
        }

    def directory_ls(self, dir_id):
        cnt_id_bytes = hash_to_bytes(dir_id)
        dir_content = map(converters.from_directory_entry,
                          self.storage.directory_ls(cnt_id_bytes))
        return list(dir_content)

    def release_get(self, rel_id):
        rel_id_bytes = hash_to_bytes(rel_id)
        rel_data = next(self.storage.release_get([rel_id_bytes]))
        return converters.from_release(rel_data)

    def revision_get(self, rev_id):
        rev_id_bytes = hash_to_bytes(rev_id)
        rev_data = next(self.storage.revision_get([rev_id_bytes]))
        return converters.from_revision(rev_data)

    def revision_log(self, rev_id, limit=None):
        rev_id_bytes = hash_to_bytes(rev_id)
        return list(map(converters.from_revision,
                        self.storage.revision_log([rev_id_bytes],
                                                  limit=limit)))

    def snapshot_get_latest(self, origin_url):
        snp = self.storage.snapshot_get_latest(origin_url)
        return converters.from_snapshot(snp)

    def origin_get(self, origin_info):
        origin = self.storage.origin_get(origin_info)
        return converters.from_origin(origin)

    def origin_visit_get(self, origin_url):
        visits = self.storage.origin_visit_get(origin_url)
        return list(map(converters.from_origin_visit, visits))

    def origin_visit_get_by(self, origin_url, visit_id):
        visit = self.storage.origin_visit_get_by(origin_url, visit_id)
        return converters.from_origin_visit(visit)

    def snapshot_get(self, snapshot_id):
        snp = self.storage.snapshot_get(hash_to_bytes(snapshot_id))
        return converters.from_snapshot(snp)

    def snapshot_get_branches(self, snapshot_id, branches_from='',
                              branches_count=1000, target_types=None):
        snp = self.storage.snapshot_get_branches(
            hash_to_bytes(snapshot_id), branches_from.encode(),
            branches_count, target_types)
        return converters.from_snapshot(snp)

    def snapshot_get_head(self, snapshot):
        if snapshot['branches']['HEAD']['target_type'] == 'alias':
            target = snapshot['branches']['HEAD']['target']
            head = snapshot['branches'][target]['target']
        else:
            head = snapshot['branches']['HEAD']['target']
        return head


class _IndexerData:
    """
    Helper class to manage indexer tests data

    It is initialized with a reference to an in-memory indexer storage
    containing raw tests data.

    It also defines class methods to retrieve those tests data in
    a json serializable format in order to ease tests implementation.

    """

    def __init__(self, tests_data):
        self.idx_storage = tests_data['idx_storage']
        self.mimetype_indexer = tests_data['mimetype_indexer']
        self.license_indexer = tests_data['license_indexer']
        self.ctags_indexer = tests_data['ctags_indexer']

    def content_add_mimetype(self, cnt_id):
        self.mimetype_indexer.run([hash_to_bytes(cnt_id)],
                                  'update-dups')

    def content_get_mimetype(self, cnt_id):
        mimetype = next(self.idx_storage.content_mimetype_get(
                        [hash_to_bytes(cnt_id)]))
        return converters.from_filetype(mimetype)

    def content_add_language(self, cnt_id):
        raise NotImplementedError('Language indexer is disabled.')
        self.language_indexer.run([hash_to_bytes(cnt_id)],
                                  'update-dups')

    def content_get_language(self, cnt_id):
        lang = next(self.idx_storage.content_language_get(
                    [hash_to_bytes(cnt_id)]))
        return converters.from_swh(lang, hashess={'id'})

    def content_add_license(self, cnt_id):
        self.license_indexer.run([hash_to_bytes(cnt_id)],
                                 'update-dups')

    def content_get_license(self, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        lic = next(self.idx_storage.content_fossology_license_get(
                   [cnt_id_bytes]))
        return converters.from_swh({'id': cnt_id_bytes,
                                    'facts': lic[cnt_id_bytes]},
                                   hashess={'id'})

    def content_add_ctags(self, cnt_id):
        self.ctags_indexer.run([hash_to_bytes(cnt_id)],
                               'update-dups')

    def content_get_ctags(self, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        ctags = self.idx_storage.content_ctags_get([cnt_id_bytes])
        for ctag in ctags:
            yield converters.from_swh(ctag, hashess={'id'})



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.utils.html import escape

from hypothesis import given

from swh.web.browse.utils import (
    get_mimetype_and_encoding_for_content, prepare_content_for_display,
    _re_encode_content
)
from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import reverse, get_swh_persistent_id
from swh.web.common.utils import gen_path_info
from swh.web.tests.django_asserts import (
    assert_contains, assert_not_contains, assert_template_used
)
from swh.web.tests.strategies import (
    content, content_text_non_utf8, content_text_no_highlight,
    content_image_type, content_text, invalid_sha1, unknown_content,
    content_utf8_detected_as_binary
)


@given(content_text())
def test_content_view_text(client, archive_data, content):
    sha1_git = content['sha1_git']

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']},
                  query_params={'path': content['path']})

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

    resp = client.get(url)

    content_display = _process_content_for_display(archive_data, content)
    mimetype = content_display['mimetype']

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    if mimetype.startswith('text/'):
        assert_contains(resp, '<code class="%s">' %
                        content_display['language'])
        assert_contains(resp, escape(content_display['content_data']))
    assert_contains(resp, url_raw)

    swh_cnt_id = get_swh_persistent_id('content', sha1_git)
    swh_cnt_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_cnt_id})
    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)


@given(content_text_no_highlight())
def test_content_view_text_no_highlight(client, archive_data, content):
    sha1_git = content['sha1_git']

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']})

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

    resp = client.get(url)

    content_display = _process_content_for_display(archive_data, content)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    assert_contains(resp, '<code class="nohighlight">')
    assert_contains(resp, escape(content_display['content_data']))
    assert_contains(resp, url_raw)

    swh_cnt_id = get_swh_persistent_id('content', sha1_git)
    swh_cnt_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_cnt_id})

    assert_contains(resp, swh_cnt_id)
    assert_contains(resp, swh_cnt_id_url)


@given(content_text_non_utf8())
def test_content_view_no_utf8_text(client, archive_data, content):
    sha1_git = content['sha1_git']

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']})

    resp = client.get(url)

    content_display = _process_content_for_display(archive_data, content)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')
    swh_cnt_id = get_swh_persistent_id('content', sha1_git)
    swh_cnt_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_cnt_id})
    assert_contains(resp, swh_cnt_id_url)
    assert_contains(resp, escape(content_display['content_data']))


@given(content_image_type())
def test_content_view_image(client, archive_data, content):
    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']})

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

    resp = client.get(url)

    content_display = _process_content_for_display(archive_data, content)
    mimetype = content_display['mimetype']
    content_data = content_display['content_data']

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')
    assert_contains(resp, '<img src="data:%s;base64,%s"/>'
                    % (mimetype, content_data))
    assert_contains(resp, url_raw)


@given(content_text())
def test_content_view_text_with_path(client, archive_data, content):
    path = content['path']

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']},
                  query_params={'path': path})

    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    assert_contains(resp, '<nav class="bread-crumbs')

    content_display = _process_content_for_display(archive_data, content)
    mimetype = content_display['mimetype']

    if mimetype.startswith('text/'):
        hljs_language = content['hljs_language']
        assert_contains(resp, '<code class="%s">' % hljs_language)
        assert_contains(resp, escape(content_display['content_data']))

    split_path = path.split('/')

    root_dir_sha1 = split_path[0]
    filename = split_path[-1]
    path = path.replace(root_dir_sha1 + '/', '').replace(filename, '')

    path_info = gen_path_info(path)

    root_dir_url = reverse('browse-directory',
                           url_args={'sha1_git': root_dir_sha1})

    assert_contains(resp, '<li class="swh-path">',
                    count=len(path_info)+1)

    assert_contains(resp, '<a href="' + root_dir_url + '">' +
                    root_dir_sha1[:7] + '</a>')

    for p in path_info:
        dir_url = reverse('browse-directory',
                          url_args={'sha1_git': root_dir_sha1,
                                    'path': p['path']})
        assert_contains(resp, '<a href="' + dir_url + '">' +
                        p['name'] + '</a>')

    assert_contains(resp, '<li>' + filename + '</li>')

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']},
                      query_params={'filename': filename})
    assert_contains(resp, url_raw)

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']},
                  query_params={'path': filename})

    resp = client.get(url)
    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    assert_not_contains(resp, '<nav class="bread-crumbs')

    invalid_path = '%s/foo/bar/baz' % root_dir_sha1
    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']},
                  query_params={'path': invalid_path})
    resp = client.get(url)
    assert resp.status_code == 404


@given(content_text())
def test_content_raw_text(client, archive_data, content):
    url = reverse('browse-content-raw',
                  url_args={'query_string': content['sha1']})

    resp = client.get(url)

    content_data = archive_data.content_get(content['sha1'])['data']

    assert resp.status_code == 200
    assert resp['Content-Type'] == 'text/plain'
    assert resp['Content-disposition'] == ('filename=%s_%s' %
                                           ('sha1', content['sha1']))
    assert resp.content == content_data

    filename = content['path'].split('/')[-1]

    url = reverse('browse-content-raw',
                  url_args={'query_string': content['sha1']},
                  query_params={'filename': filename})

    resp = client.get(url)

    assert resp.status_code == 200
    assert resp['Content-Type'] == 'text/plain'
    assert resp['Content-disposition'] == 'filename=%s' % filename
    assert resp.content == content_data


@given(content_text_non_utf8())
def test_content_raw_no_utf8_text(client, content):
    url = reverse('browse-content-raw',
                  url_args={'query_string': content['sha1']})

    resp = client.get(url)
    assert resp.status_code == 200
    _, encoding = get_mimetype_and_encoding_for_content(resp.content)
    assert encoding == content['encoding']


@given(content_image_type())
def test_content_raw_bin(client, archive_data, content):
    url = reverse('browse-content-raw',
                  url_args={'query_string': content['sha1']})

    resp = client.get(url)

    filename = content['path'].split('/')[-1]
    content_data = archive_data.content_get(content['sha1'])['data']

    assert resp.status_code == 200
    assert resp['Content-Type'] == 'application/octet-stream'
    assert resp['Content-disposition'] == \
        'attachment; filename=%s_%s' % \
        ('sha1', content['sha1'])
    assert resp.content == content_data

    url = reverse('browse-content-raw',
                  url_args={'query_string': content['sha1']},
                  query_params={'filename': filename})

    resp = client.get(url)

    assert resp.status_code == 200
    assert resp['Content-Type'] == 'application/octet-stream'
    assert resp['Content-disposition'] == \
        'attachment; filename=%s' % filename
    assert resp.content == content_data


@given(invalid_sha1(), unknown_content())
def test_content_request_errors(client, invalid_sha1, unknown_content):
    url = reverse('browse-content',
                  url_args={'query_string': invalid_sha1})
    resp = client.get(url)
    assert resp.status_code == 400
    assert_template_used(resp, 'error.html')

    url = reverse('browse-content',
                  url_args={'query_string': unknown_content['sha1']})
    resp = client.get(url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')


@given(content())
def test_content_bytes_missing(client, archive_data, mocker, content):
    mock_service = mocker.patch('swh.web.browse.utils.service')
    content_data = archive_data.content_get_metadata(content['sha1'])
    content_data['data'] = None

    mock_service.lookup_content.return_value = content_data
    mock_service.lookup_content_filetype.side_effect = Exception()
    mock_service.lookup_content_raw.side_effect = NotFoundExc(
        'Content bytes not available!')

    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']})

    resp = client.get(url)

    assert resp.status_code == 404
    assert_template_used(resp, 'browse/content.html')


def test_content_too_large(client, mocker):
    mock_request_content = mocker.patch(
        'swh.web.browse.views.content.request_content')
    stub_content_too_large_data = {
        'checksums': {
            'sha1': '8624bcdae55baeef00cd11d5dfcfa60f68710a02',
            'sha1_git': '94a9ed024d3859793618152ea559a168bbcbb5e2',
            'sha256': ('8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b16'
                        '5a1dcd80c7c545eb65b903'),
            'blake2s256': ('38702b7168c7785bfe748b51b45d9856070ba90'
                           'f9dc6d90f2ea75d4356411ffe')
        },
        'length': 30000000,
        'raw_data': None,
        'mimetype': 'text/plain',
        'encoding': 'us-ascii',
        'language': 'not detected',
        'licenses': 'GPL',
        'error_code': 200,
        'error_message': '',
        'error_description': ''
    }

    content_sha1 = stub_content_too_large_data['checksums']['sha1']

    mock_request_content.return_value = stub_content_too_large_data

    url = reverse('browse-content',
                  url_args={'query_string': content_sha1})

    url_raw = reverse('browse-content-raw',
                      url_args={'query_string': content_sha1})

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/content.html')

    assert_contains(resp, 'Content is too large to be displayed')
    assert_contains(resp, url_raw)


@given(content())
def test_content_uppercase(client, content):
    url = reverse('browse-content-uppercase-checksum',
                  url_args={'query_string': content['sha1'].upper()})
    resp = client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('browse-content',
                           url_args={'query_string': content['sha1']})

    assert resp['location'] == redirect_url


@given(content_utf8_detected_as_binary())
def test_content_utf8_detected_as_binary_display(client, archive_data,
                                                 content):
    url = reverse('browse-content',
                  url_args={'query_string': content['sha1']})
    resp = client.get(url)

    content_display = _process_content_for_display(archive_data, content)

    assert_contains(resp, escape(content_display['content_data']))


def _process_content_for_display(archive_data, content):
    content_data = archive_data.content_get(content['sha1'])

    mime_type, encoding = get_mimetype_and_encoding_for_content(
        content_data['data'])

    mime_type, encoding, content_data = _re_encode_content(
        mime_type, encoding, content_data['data'])

    content_display = prepare_content_for_display(content_data, mime_type,
                                                  content['path'])

    assert type(content_display['content_data']) == str

    return content_display



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import docutils.nodes
import docutils.parsers.rst
import docutils.utils
import functools
from functools import wraps
import os
import re
import textwrap

from rest_framework.decorators import api_view
import sentry_sdk

from swh.web.common.utils import parse_rst
from swh.web.api.apiurls import APIUrls
from swh.web.api.apiresponse import make_api_response, error_response


class _HTTPDomainDocVisitor(docutils.nodes.NodeVisitor):
    """
    docutils visitor for walking on a parsed rst document containing sphinx
    httpdomain roles. Its purpose is to extract relevant info regarding swh
    api endpoints (for instance url arguments) from their docstring written
    using sphinx httpdomain.
    """

    # httpdomain roles we want to parse (based on sphinxcontrib.httpdomain 1.6)
    parameter_roles = ('param', 'parameter', 'arg', 'argument')

    response_json_object_roles = ('resjsonobj', 'resjson', '>jsonobj', '>json')

    response_json_array_roles = ('resjsonarr', '>jsonarr')

    query_parameter_roles = ('queryparameter', 'queryparam', 'qparam', 'query')

    request_header_roles = ('<header', 'reqheader', 'requestheader')

    response_header_roles = ('>header', 'resheader', 'responseheader')

    status_code_roles = ('statuscode', 'status', 'code')

    def __init__(self, document, urls, data):
        super().__init__(document)
        self.urls = urls
        self.url_idx = 0
        self.data = data
        self.args_set = set()
        self.params_set = set()
        self.returns_set = set()
        self.status_codes_set = set()
        self.reqheaders_set = set()
        self.resheaders_set = set()
        self.field_list_visited = False

    def process_paragraph(self, par):
        """
        Process extracted paragraph text before display.
        Cleanup document model markups and transform the
        paragraph into a valid raw rst string (as the apidoc
        documentation transform rst to html when rendering).
        """
        par = par.replace('\n', ' ')
        # keep emphasized, strong and literal text
        par = par.replace('<emphasis>', '*')
        par = par.replace('</emphasis>', '*')
        par = par.replace('<strong>', '**')
        par = par.replace('</strong>', '**')
        par = par.replace('<literal>', '``')
        par = par.replace('</literal>', '``')
        # remove parsed document markups
        par = re.sub('<[^<]+?>', '', par)
        # api urls cleanup to generate valid links afterwards
        par = re.sub(r'\(\w+\)', '', par)
        par = re.sub(r'\[.*\]', '', par)
        par = par.replace('//', '/')
        # transform references to api endpoints into valid rst links
        par = re.sub(':http:get:`([^,]*)`', r'`<\1>`_', par)
        # transform references to some elements into bold text
        par = re.sub(':http:header:`(.*)`', r'**\1**', par)
        par = re.sub(':func:`(.*)`', r'**\1**', par)
        return par

    def visit_field_list(self, node):
        """
        Visit parsed rst field lists to extract relevant info
        regarding api endpoint.
        """
        self.field_list_visited = True
        for child in node.traverse():
            # get the parsed field name
            if isinstance(child, docutils.nodes.field_name):
                field_name = child.astext()
            # parse field text
            elif isinstance(child, docutils.nodes.paragraph):
                text = self.process_paragraph(str(child))
                field_data = field_name.split(' ')
                # Parameters
                if field_data[0] in self.parameter_roles:
                    if field_data[2] not in self.args_set:
                        self.data['args'].append({'name': field_data[2],
                                                  'type': field_data[1],
                                                  'doc': text})
                        self.args_set.add(field_data[2])
                # Query Parameters
                if field_data[0] in self.query_parameter_roles:
                    if field_data[2] not in self.params_set:
                        self.data['params'].append({'name': field_data[2],
                                                    'type': field_data[1],
                                                    'doc': text})
                        self.params_set.add(field_data[2])
                # Response type
                if field_data[0] in self.response_json_array_roles or \
                        field_data[0] in self.response_json_object_roles:
                    # array
                    if field_data[0] in self.response_json_array_roles:
                        self.data['return_type'] = 'array'
                    # object
                    else:
                        self.data['return_type'] = 'object'
                    # returned object field
                    if field_data[2] not in self.returns_set:
                        self.data['returns'].append({'name': field_data[2],
                                                     'type': field_data[1],
                                                     'doc': text})
                        self.returns_set.add(field_data[2])
                # Status Codes
                if field_data[0] in self.status_code_roles:
                    if field_data[1] not in self.status_codes_set:
                        self.data['status_codes'].append({'code': field_data[1], # noqa
                                                          'doc': text})
                        self.status_codes_set.add(field_data[1])
                # Request Headers
                if field_data[0] in self.request_header_roles:
                    if field_data[1] not in self.reqheaders_set:
                        self.data['reqheaders'].append({'name': field_data[1],
                                                        'doc': text})
                        self.reqheaders_set.add(field_data[1])
                # Response Headers
                if field_data[0] in self.response_header_roles:
                    if field_data[1] not in self.resheaders_set:
                        resheader = {'name': field_data[1],
                                     'doc': text}
                        self.data['resheaders'].append(resheader)
                        self.resheaders_set.add(field_data[1])
                        if resheader['name'] == 'Content-Type' and \
                                resheader['doc'] == 'application/octet-stream':
                            self.data['return_type'] = 'octet stream'

    def visit_paragraph(self, node):
        """
        Visit relevant paragraphs to parse
        """
        # only parsed top level paragraphs
        if isinstance(node.parent, docutils.nodes.block_quote):
            text = self.process_paragraph(str(node))
            # endpoint description
            if (not text.startswith('**') and
                    text not in self.data['description']):
                self.data['description'] += '\n\n' if self.data['description'] else '' # noqa
                self.data['description'] += text
            # http methods
            elif text.startswith('**Allowed HTTP Methods:**'):
                text = text.replace('**Allowed HTTP Methods:**', '')
                http_methods = text.strip().split(',')
                http_methods = [m[m.find('`')+1:-1].upper()
                                for m in http_methods]
                self.data['urls'].append({'rule': self.urls[self.url_idx],
                                          'methods': http_methods})
                self.url_idx += 1

    def visit_literal_block(self, node):
        """
        Visit literal blocks
        """
        text = node.astext()
        # literal block in endpoint description
        if not self.field_list_visited:
            self.data['description'] += \
                ':\n\n%s\n' % textwrap.indent(text, '\t')
        # extract example url
        if ':swh_web_api:' in text:
            self.data['examples'].append(
                '/api/1/' + re.sub('.*`(.*)`.*', r'\1', text))

    def visit_bullet_list(self, node):
        # bullet list in endpoint description
        if not self.field_list_visited:
            self.data['description'] += '\n\n'
            for child in node.traverse():
                # process list item
                if isinstance(child, docutils.nodes.paragraph):
                    line_text = self.process_paragraph(str(child))
                    self.data['description'] += '\t* %s\n' % line_text

    def visit_warning(self, node):
        text = self.process_paragraph(str(node))
        rst_warning = '\n\n.. warning::\n%s\n' % textwrap.indent(text, '\t')
        if rst_warning not in self.data['description']:
            self.data['description'] += rst_warning

    def unknown_visit(self, node):
        pass

    def depart_document(self, node):
        """
        End of parsing extra processing
        """
        default_methods = ['GET', 'HEAD', 'OPTIONS']
        # ensure urls info is present and set default http methods
        if not self.data['urls']:
            for url in self.urls:
                self.data['urls'].append({'rule': url,
                                          'methods': default_methods})

    def unknown_departure(self, node):
        pass


def _parse_httpdomain_doc(doc, data):
    doc_lines = doc.split('\n')
    doc_lines_filtered = []
    urls = []
    # httpdomain is a sphinx extension that is unknown to docutils but
    # fortunately we can still parse its directives' content,
    # so remove lines with httpdomain directives before executing the
    # rst parser from docutils
    for doc_line in doc_lines:
        if '.. http' not in doc_line:
            doc_lines_filtered.append(doc_line)
        else:
            url = doc_line[doc_line.find('/'):]
            # emphasize url arguments for html rendering
            url = re.sub(r'\((\w+)\)', r' **\(\1\)** ', url)
            urls.append(url)
    # parse the rst docstring and do not print system messages about
    # unknown httpdomain roles
    document = parse_rst('\n'.join(doc_lines_filtered), report_level=5)
    # remove the system_message nodes from the parsed document
    for node in document.traverse(docutils.nodes.system_message):
        node.parent.remove(node)
    # visit the document nodes to extract relevant endpoint info
    visitor = _HTTPDomainDocVisitor(document, urls, data)
    document.walkabout(visitor)


class APIDocException(Exception):
    """
    Custom exception to signal errors in the use of the APIDoc decorators
    """


def api_doc(route, noargs=False, need_params=False, tags=[],
            handle_response=False, api_version='1'):
    """
    Decorate an API function to register it in the API doc route index
    and create the corresponding DRF route.

    Args:
        route (str): documentation page's route
        noargs (boolean): set to True if the route has no arguments, and its
            result should be displayed anytime its documentation
            is requested. Default to False
        need_params (boolean): specify the route requires query parameters
            otherwise errors will occur. It enables to avoid displaying the
            invalid response in its HTML documentation. Default to False.
        tags (list): Further information on api endpoints. Two values are
            possibly expected:

                * hidden: remove the entry points from the listing
                * upcoming: display the entry point but it is not followable

        handle_response (boolean): indicate if the decorated function takes
            care of creating the HTTP response or delegates that task to the
            apiresponse module
        api_version (str): api version string

    """
    urlpattern = '^' + api_version + route + '$'
    tags = set(tags)

    # @api_doc() Decorator call
    def decorator(f):

        # If the route is not hidden, add it to the index
        if 'hidden' not in tags:
            doc_data = get_doc_data(f, route, noargs)
            doc_desc = doc_data['description']
            first_dot_pos = doc_desc.find('.')
            APIUrls.add_route(route, doc_desc[:first_dot_pos+1],
                              tags=tags)

        # If the decorated route has arguments, we create a specific
        # documentation view
        if not noargs:

            @api_view(['GET', 'HEAD'])
            @wraps(f)
            def doc_view(request):
                doc_data = get_doc_data(f, route, noargs)
                return make_api_response(request, None, doc_data)

            view_name = 'api-%s-%s' % \
                (api_version, route[1:-1].replace('/', '-'))
            APIUrls.add_url_pattern(urlpattern, doc_view, view_name)

        @wraps(f)
        def documented_view(request, **kwargs):
            doc_data = get_doc_data(f, route, noargs)

            try:
                response = f(request, **kwargs)
            except Exception as exc:
                sentry_sdk.capture_exception(exc)
                if request.accepted_media_type == 'text/html' and \
                        need_params and not request.query_params:
                    response = None
                else:
                    return error_response(request, exc, doc_data)

            if handle_response:
                return response
            else:
                return make_api_response(request, response, doc_data)

        return documented_view

    return decorator


@functools.lru_cache(maxsize=32)
def get_doc_data(f, route, noargs):
    """
    Build documentation data for the decorated api endpoint function
    """
    data = {
        'description': '',
        'response_data': None,
        'urls': [],
        'args': [],
        'params': [],
        'resheaders': [],
        'reqheaders': [],
        'return_type': '',
        'returns': [],
        'status_codes': [],
        'examples': [],
        'route': route,
        'noargs': noargs
    }

    if not f.__doc__:
        raise APIDocException('apidoc: expected a docstring'
                              ' for function %s'
                              % (f.__name__,))

    # use raw docstring as endpoint documentation if sphinx
    # httpdomain is not used
    if '.. http' not in f.__doc__:
        data['description'] = f.__doc__
    # else parse the sphinx httpdomain docstring with docutils
    # (except when building the swh-web documentation through autodoc
    # sphinx extension, not needed and raise errors with sphinx >= 1.7)
    elif 'SWH_WEB_DOC_BUILD' not in os.environ:
        _parse_httpdomain_doc(f.__doc__, data)
        # process returned object info for nicer html display
        returns_list = ''
        for ret in data['returns']:
            returns_list += '\t* **%s (%s)**: %s\n' %\
                (ret['name'], ret['type'], ret['doc'])
        data['returns_list'] = returns_list

    return data


DOC_COMMON_HEADERS = '''
        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept`
            header of request'''
DOC_RESHEADER_LINK = '''
        :resheader Link: indicates that a subsequent result page is
            available and contains the url pointing to it
'''

DEFAULT_SUBSTITUTIONS = {
    'common_headers': DOC_COMMON_HEADERS,
    'resheader_link': DOC_RESHEADER_LINK,
}


def format_docstring(**substitutions):
    def decorator(f):
        f.__doc__ = f.__doc__.format(**{
            **DEFAULT_SUBSTITUTIONS, **substitutions})
        return f
    return decorator



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.http import HttpResponse

from swh.web.common import service
from swh.web.common.utils import reverse
from swh.web.api import utils
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup


DOC_RETURN_REVISION = '''
        :>json object author: information about the author of the revision
        :>json object committer: information about the committer of the
            revision
        :>json string committer_date: ISO representation of the commit date
            (in UTC)
        :>json string date: ISO representation of the revision date (in UTC)
        :>json string directory: the unique identifier that revision points to
        :>json string directory_url: link to
            :http:get:`/api/1/directory/(sha1_git)/[(path)/]` to get
            information about the directory associated to the revision
        :>json string id: the revision unique identifier
        :>json boolean merge: whether or not the revision corresponds to a
            merge commit
        :>json string message: the message associated to the revision
        :>json array parents: the parents of the revision, i.e. the previous
            revisions that head directly to it, each entry of that array
            contains an unique parent revision identifier but also a link to
            :http:get:`/api/1/revision/(sha1_git)/` to get more information
            about it
        :>json string type: the type of the revision
''' # noqa

DOC_RETURN_REVISION_ARRAY = \
    DOC_RETURN_REVISION.replace(':>json', ':>jsonarr')


def _revision_directory_by(revision, path, request_path,
                           limit=100, with_data=False):
    """
    Compute the revision matching criterion's directory or content data.

    Args:
        revision: dictionary of criterions representing a revision to lookup
        path: directory's path to lookup
        request_path: request path which holds the original context to
        limit: optional query parameter to limit the revisions log
        (default to 100). For now, note that this limit could impede the
        transitivity conclusion about sha1_git not being an ancestor of
        with_data: indicate to retrieve the content's raw data if path resolves
        to a content.

    """
    def enrich_directory_local(dir, context_url=request_path):
        return utils.enrich_directory(dir, context_url)

    rev_id, result = service.lookup_directory_through_revision(
        revision, path, limit=limit, with_data=with_data)

    content = result['content']
    if result['type'] == 'dir':  # dir_entries
        result['content'] = list(map(enrich_directory_local, content))
    elif result['type'] == 'file':  # content
        result['content'] = utils.enrich_content(content)
    elif result['type'] == 'rev':  # revision
        result['content'] = utils.enrich_revision(content)

    return result


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/', 'api-1-revision',
           checksum_args=['sha1_git'])
@api_doc('/revision/')
@format_docstring(return_revision=DOC_RETURN_REVISION)
def api_revision(request, sha1_git):
    """
    .. http:get:: /api/1/revision/(sha1_git)/

        Get information about a revision in the archive. Revisions are
        identified by **sha1** checksums, compatible with Git commit
        identifiers.
        See :func:`swh.model.identifiers.revision_identifier` in our data model
        module for details about how they are computed.

        :param string sha1_git: hexadecimal representation of the revision
            **sha1_git** identifier

        {common_headers}

        {return_revision}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/`
    """ # noqa
    return api_lookup(
        service.lookup_revision, sha1_git,
        notfound_msg='Revision with sha1_git {} not found.'.format(sha1_git),
        enrich_fn=utils.enrich_revision,
        request=request)


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/raw/',
           'api-1-revision-raw-message', checksum_args=['sha1_git'])
@api_doc('/revision/raw/', tags=['hidden'], handle_response=True)
def api_revision_raw_message(request, sha1_git):
    """Return the raw data of the message of revision identified by sha1_git
    """
    raw = service.lookup_revision_message(sha1_git)
    response = HttpResponse(raw['message'],
                            content_type='application/octet-stream')
    response['Content-disposition'] = \
        'attachment;filename=rev_%s_raw' % sha1_git
    return response


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/directory/',
           'api-1-revision-directory', checksum_args=['sha1_git'])
@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/directory/(?P<dir_path>.+)/',
           'api-1-revision-directory', checksum_args=['sha1_git'])
@api_doc('/revision/directory/')
@format_docstring()
def api_revision_directory(request, sha1_git,
                           dir_path=None,
                           with_data=False):
    """
    .. http:get:: /api/1/revision/(sha1_git)/directory/[(path)/]

        Get information about directory (entry) objects associated to revisions.
        Each revision is associated to a single "root" directory.
        This endpoint behaves like :http:get:`/api/1/directory/(sha1_git)/[(path)/]`,
        but operates on the root directory associated to a given revision.

        :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier
        :param string path: optional parameter to get information about the directory entry
            pointed by that relative path

        {common_headers}

        :>json array content: directory entries as returned by :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
        :>json string path: path of directory from the revision root one
        :>json string revision: the unique revision identifier
        :>json string type: the type of the directory

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/f1b94134a4b879bc55c3dacdb496690c8ebdc03f/directory/`
    """ # noqa
    return _revision_directory_by({'sha1_git': sha1_git},
                                  dir_path, request.path,
                                  with_data=with_data)


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/log/', 'api-1-revision-log',
           checksum_args=['sha1_git'])
@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)'
           r'/prev/(?P<prev_sha1s>[0-9a-f]*/*)/log/',
           'api-1-revision-log', checksum_args=['sha1_git', 'prev_sha1s'])
@api_doc('/revision/log/')
@format_docstring(return_revision_array=DOC_RETURN_REVISION_ARRAY)
def api_revision_log(request, sha1_git, prev_sha1s=None):
    """
    .. http:get:: /api/1/revision/(sha1_git)[/prev/(prev_sha1s)]/log/

        Get a list of all revisions heading to a given one, in other words show the commit log.

        :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier
        :param string prev_sha1s: optional parameter representing the navigation breadcrumbs
            (descendant revisions previously visited). If multiple values, use / as delimiter.
            If provided, revisions information will be added at the beginning of the returned list.
        :query int per_page: number of elements in the returned list, for pagination purpose

        {common_headers}
        {resheader_link}

        {return_revision_array}

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/e1a315fa3fa734e2a6154ed7b5b9ae0eb8987aad/log/`
    """ # noqa
    result = {}
    per_page = int(request.query_params.get('per_page', '10'))

    def lookup_revision_log_with_limit(s, limit=per_page+1):
        return service.lookup_revision_log(s, limit)

    error_msg = 'Revision with sha1_git %s not found.' % sha1_git
    rev_get = api_lookup(lookup_revision_log_with_limit, sha1_git,
                         notfound_msg=error_msg,
                         enrich_fn=utils.enrich_revision,
                         request=request)

    nb_rev = len(rev_get)
    if nb_rev == per_page+1:
        rev_backward = rev_get[:-1]
        new_last_sha1 = rev_get[-1]['id']
        query_params = {}

        if request.query_params.get('per_page'):
            query_params['per_page'] = per_page

        result['headers'] = {
            'link-next': reverse('api-1-revision-log',
                                 url_args={'sha1_git': new_last_sha1},
                                 query_params=query_params,
                                 request=request)
        }

    else:
        rev_backward = rev_get

    if not prev_sha1s:  # no nav breadcrumbs, so we're done
        revisions = rev_backward

    else:
        rev_forward_ids = prev_sha1s.split('/')
        rev_forward = api_lookup(
            service.lookup_revision_multiple, rev_forward_ids,
            notfound_msg=error_msg,
            enrich_fn=utils.enrich_revision,
            request=request)
        revisions = rev_forward + rev_backward

    result.update({
        'results': revisions
    })
    return result



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import traceback

from django.http import HttpResponse
from django.shortcuts import render
from django.utils.safestring import mark_safe
from django.utils.html import escape
import sentry_sdk

from swh.web.config import get_config


class BadInputExc(ValueError):
    """Wrong request to the api.

    Example: Asking a content with the wrong identifier format.

    """
    pass


class NotFoundExc(Exception):
    """Good request to the api but no result were found.

    Example: Asking a content with the right identifier format but
    that content does not exist.

    """
    pass


class ForbiddenExc(Exception):
    """Good request to the api, forbidden result to return due to enforce
       policy.

    Example: Asking for a raw content which exists but whose mimetype
    is not text.

    """
    pass


http_status_code_message = {
    400: 'Bad Request',
    401: 'Unauthorized',
    403: 'Access Denied',
    404: 'Resource not found',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service unavailable'
}


def _generate_error_page(request, error_code, error_description):
    return render(request, 'error.html',
                  {'error_code': error_code,
                   'error_message': http_status_code_message[error_code],
                   'error_description': mark_safe(error_description)},
                  status=error_code)


def swh_handle400(request, exception=None):
    """
    Custom Django HTTP error 400 handler for swh-web.
    """
    error_description = ('The server cannot process the request to %s due to '
                         'something that is perceived to be a client error.' %
                         escape(request.META['PATH_INFO']))
    return _generate_error_page(request, 400, error_description)


def swh_handle403(request, exception=None):
    """
    Custom Django HTTP error 403 handler for swh-web.
    """
    error_description = ('The resource %s requires an authentication.' %
                         escape(request.META['PATH_INFO']))
    return _generate_error_page(request, 403, error_description)


def swh_handle404(request, exception=None):
    """
    Custom Django HTTP error 404 handler for swh-web.
    """
    error_description = ('The resource %s could not be found on the server.' %
                         escape(request.META['PATH_INFO']))
    return _generate_error_page(request, 404, error_description)


def swh_handle500(request):
    """
    Custom Django HTTP error 500 handler for swh-web.
    """
    error_description = ('An unexpected condition was encountered when '
                         'requesting resource %s.' %
                         escape(request.META['PATH_INFO']))
    return _generate_error_page(request, 500, error_description)


def handle_view_exception(request, exc, html_response=True):
    """
    Function used to generate an error page when an exception
    was raised inside a swh-web browse view.
    """
    sentry_sdk.capture_exception(exc)
    error_code = 500
    error_description = '%s: %s' % (type(exc).__name__, str(exc))
    if get_config()['debug']:
        error_description = traceback.format_exc()
    if isinstance(exc, BadInputExc):
        error_code = 400
    if isinstance(exc, ForbiddenExc):
        error_code = 403
    if isinstance(exc, NotFoundExc):
        error_code = 404
    if html_response:
        return _generate_error_page(request, error_code, error_description)
    else:
        return HttpResponse(error_description, content_type='text/plain',
                            status=error_code)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import pytest

from swh.web.common import utils
from swh.web.common.exc import BadInputExc


def test_shorten_path_noop():
    noops = [
        '/api/',
        '/browse/',
        '/content/symbol/foobar/'
    ]

    for noop in noops:
        assert utils.shorten_path(noop) == noop


def test_shorten_path_sha1():
    sha1 = 'aafb16d69fd30ff58afdd69036a26047f3aebdc6'
    short_sha1 = sha1[:8] + '...'

    templates = [
        '/api/1/content/sha1:%s/',
        '/api/1/content/sha1_git:%s/',
        '/api/1/directory/%s/',
        '/api/1/content/sha1:%s/ctags/',
    ]

    for template in templates:
        assert utils.shorten_path(template % sha1) == template % short_sha1


def test_shorten_path_sha256():
    sha256 = ('aafb16d69fd30ff58afdd69036a26047'
              '213add102934013a014dfca031c41aef')
    short_sha256 = sha256[:8] + '...'

    templates = [
        '/api/1/content/sha256:%s/',
        '/api/1/directory/%s/',
        '/api/1/content/sha256:%s/filetype/',
    ]

    for template in templates:
        assert utils.shorten_path(template % sha256) == template % short_sha256


def test_parse_timestamp():
    input_timestamps = [
        None,
        '2016-01-12',
        '2016-01-12T09:19:12+0100',
        'Today is January 1, 2047 at 8:21:00AM',
        '1452591542',
    ]

    output_dates = [
        None,
        datetime.datetime(2016, 1, 12, 0, 0),
        datetime.datetime(2016, 1, 12, 8, 19, 12,
                          tzinfo=datetime.timezone.utc),
        datetime.datetime(2047, 1, 1, 8, 21),
        datetime.datetime(2016, 1, 12, 9, 39, 2,
                          tzinfo=datetime.timezone.utc),
    ]

    for ts, exp_date in zip(input_timestamps, output_dates):
        assert utils.parse_timestamp(ts) == exp_date


def test_format_utc_iso_date():
    assert (utils.format_utc_iso_date('2017-05-04T13:27:13+02:00') ==
            '04 May 2017, 11:27 UTC')


def test_gen_path_info():
    input_path = '/home/user/swh-environment/swh-web/'
    expected_result = [
        {'name': 'home', 'path': 'home'},
        {'name': 'user', 'path': 'home/user'},
        {'name': 'swh-environment', 'path': 'home/user/swh-environment'},
        {'name': 'swh-web', 'path': 'home/user/swh-environment/swh-web'}
    ]
    path_info = utils.gen_path_info(input_path)
    assert path_info == expected_result

    input_path = 'home/user/swh-environment/swh-web'
    path_info = utils.gen_path_info(input_path)
    assert path_info == expected_result


def test_get_swh_persistent_id():
    swh_object_type = 'content'
    sha1_git = 'aafb16d69fd30ff58afdd69036a26047f3aebdc6'

    expected_swh_id = 'swh:1:cnt:' + sha1_git

    assert (utils.get_swh_persistent_id(swh_object_type, sha1_git) ==
            expected_swh_id)

    with pytest.raises(BadInputExc) as e:
        utils.get_swh_persistent_id('foo', sha1_git)
    assert e.match('Invalid object')

    with pytest.raises(BadInputExc) as e:
        utils.get_swh_persistent_id(swh_object_type, 'not a valid id')
    assert e.match('Invalid object')



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.web.common.swh_templatetags import (
    urlize_links_and_mails, urlize_header_links, safe_docstring_display
)


def test_urlize_http_link():
    link = 'https://example.com/api/1/abc/'
    expected_content = f'<a href="{link}">{link}</a>'

    assert urlize_links_and_mails(link) == expected_content


def test_urlize_email():
    email = 'someone@example.com'
    expected_content = f'<a href="mailto:{email}">{email}</a>'

    assert urlize_links_and_mails(email) == expected_content


def test_urlize_header_links():

    next_link = 'https://example.com/api/1/abc/'
    prev_link = 'https://example.com/api/1/def/'

    content = f'<{next_link}>; rel="next"\n<{prev_link}>; rel="prev"'

    expected_content = (
        f'<<a href="{next_link}">{next_link}</a>>; rel="next"\n'
        f'<<a href="{prev_link}">{prev_link}</a>>; rel="prev"')

    assert urlize_header_links(content) == expected_content


# remove deprecation warnings related to docutils
@pytest.mark.filterwarnings(
    'ignore:.*U.*mode is deprecated:DeprecationWarning')
def test_safe_docstring_display():
    # update api link with html links content with links
    docstring = (
        'This is my list header:\n\n'
        '    - Here is item 1, with a continuation\n'
        '      line right here\n'
        '    - Here is item 2\n\n'
        '    Here is something that is not part of the list'
    )

    expected_docstring = (
        '<p>This is my list header:</p>\n'
        '<ul class="docstring">\n'
        '<li>Here is item 1, with a continuation\n'
        'line right here</li>\n'
        '<li>Here is item 2</li>\n'
        '</ul>\n'
        '<p>Here is something that is not part of the list</p>\n'
    )

    assert safe_docstring_display(docstring) == expected_docstring



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, Optional, Any

from django.http import HttpRequest

from swh.web.common.utils import reverse, resolve_branch_alias
from swh.web.common.query import parse_hash


def filter_field_keys(data, field_keys):
    """Given an object instance (directory or list), and a csv field keys
    to filter on.

    Return the object instance with filtered keys.

    Note: Returns obj as is if it's an instance of types not in (dictionary,
    list)

    Args:
        - data: one object (dictionary, list...) to filter.
        - field_keys: csv or set of keys to filter the object on

    Returns:
        obj filtered on field_keys

    """
    if isinstance(data, map):
        return map(lambda x: filter_field_keys(x, field_keys), data)
    if isinstance(data, list):
        return [filter_field_keys(x, field_keys) for x in data]
    if isinstance(data, dict):
        return {k: v for (k, v) in data.items() if k in field_keys}
    return data


def person_to_string(person):
    """Map a person (person, committer, tagger, etc...) to a string.

    """
    return ''.join([person['name'], ' <', person['email'], '>'])


def enrich_object(object: Dict[str, str],
                  request: Optional[HttpRequest] = None) -> Dict[str, str]:
    """Enrich an object (revision, release) with link to the 'target' of
    type 'target_type'.

    Args:
        object: An object with target and target_type keys
        (e.g. release, revision)
        request: Absolute URIs will be generated if provided

    Returns:
        Object enriched with target object url (revision, release, content,
        directory)

    """
    if 'target' in object and 'target_type' in object:
        if object['target_type'] in ('revision', 'release', 'directory'):
            object['target_url'] = reverse(
                'api-1-%s' % object['target_type'],
                url_args={'sha1_git': object['target']},
                request=request)
        elif object['target_type'] == 'content':
            object['target_url'] = reverse(
                'api-1-content',
                url_args={'q': 'sha1_git:' + object['target']},
                request=request)
        elif object['target_type'] == 'snapshot':
            object['target_url'] = reverse(
                'api-1-snapshot',
                url_args={'snapshot_id': object['target']},
                request=request)

    return object


enrich_release = enrich_object


def enrich_directory(directory: Dict[str, str],
                     request: Optional[HttpRequest] = None) -> Dict[str, str]:
    """Enrich directory with url to content or directory.

    Args:
        directory: dict of data associated to a swh directory object
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched directory dict filled with additional urls
    """
    if 'type' in directory:
        target_type = directory['type']
        target = directory['target']
        if target_type == 'file':
            directory['target_url'] = reverse(
                'api-1-content',
                url_args={'q': 'sha1_git:%s' % target},
                request=request)
        elif target_type == 'dir':
            directory['target_url'] = reverse(
                'api-1-directory',
                url_args={'sha1_git': target},
                request=request)
        else:
            directory['target_url'] = reverse(
                'api-1-revision',
                url_args={'sha1_git': target},
                request=request)

    return directory


def enrich_metadata_endpoint(content_metadata: Dict[str, str],
                             request: Optional[HttpRequest] = None
                             ) -> Dict[str, str]:
    """Enrich content metadata dict with link to the upper metadata endpoint.

    Args:
        content_metadata: dict of data associated to a swh content metadata
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched content metadata dict filled with an additional url
    """
    c = content_metadata
    c['content_url'] = reverse('api-1-content',
                               url_args={'q': 'sha1:%s' % c['id']},
                               request=request)
    return c


def enrich_content(content: Dict[str, Any],
                   top_url: Optional[bool] = False,
                   query_string: Optional[str] = None,
                   request: Optional[HttpRequest] = None) -> Dict[str, str]:
    """Enrich content with links to:
        - data_url: its raw data
        - filetype_url: its filetype information
        - language_url: its programming language information
        - license_url: its licensing information

    Args:
        content: dict of data associated to a swh content object
        top_url: whether or not to include the content url in
            the enriched data
        query_string: optional query string of type '<algo>:<hash>'
            used when requesting the content, it acts as a hint
            for picking the same hash method when computing
            the url listed above
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched content dict filled with additional urls

    """
    checksums = content
    if 'checksums' in content:
        checksums = content['checksums']
    hash_algo = 'sha1'
    if query_string:
        hash_algo = parse_hash(query_string)[0]
    if hash_algo in checksums:
        q = '%s:%s' % (hash_algo, checksums[hash_algo])
        if top_url:
            content['content_url'] = reverse(
                'api-1-content', url_args={'q': q})
        content['data_url'] = reverse('api-1-content-raw',
                                      url_args={'q': q},
                                      request=request)
        content['filetype_url'] = reverse(
            'api-1-content-filetype',
            url_args={'q': q},
            request=request)
        content['language_url'] = reverse(
            'api-1-content-language',
            url_args={'q': q},
            request=request)
        content['license_url'] = reverse(
            'api-1-content-license',
            url_args={'q': q},
            request=request)

    return content


def enrich_revision(revision: Dict[str, Any],
                    request: Optional[HttpRequest] = None) -> Dict[str, Any]:
    """Enrich revision with links where it makes sense (directory, parents).
    Keep track of the navigation breadcrumbs if they are specified.

    Args:
        revision: the revision as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched revision dict filled with additional urls
    """

    revision['url'] = reverse('api-1-revision',
                              url_args={'sha1_git': revision['id']},
                              request=request)
    revision['history_url'] = reverse('api-1-revision-log',
                                      url_args={'sha1_git': revision['id']},
                                      request=request)

    if 'directory' in revision:
        revision['directory_url'] = reverse(
            'api-1-directory',
            url_args={'sha1_git': revision['directory']},
            request=request)

    if 'parents' in revision:
        parents = []
        for parent in revision['parents']:
            parents.append({
                'id': parent,
                'url': reverse('api-1-revision',
                               url_args={'sha1_git': parent},
                               request=request)
            })

        revision['parents'] = parents

    if 'children' in revision:
        children = []
        for child in revision['children']:
            children.append(reverse(
                'api-1-revision',
                url_args={'sha1_git': child},
                request=request))
        revision['children_urls'] = children

    if 'message_decoding_failed' in revision:
        revision['message_url'] = reverse(
            'api-1-revision-raw-message',
            url_args={'sha1_git': revision['id']},
            request=request)

    return revision


def enrich_snapshot(snapshot: Dict[str, Any],
                    request: Optional[HttpRequest] = None) -> Dict[str, Any]:
    """Enrich snapshot with links to the branch targets

    Args:
        snapshot: the snapshot as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched snapshot dict filled with additional urls
    """
    if 'branches' in snapshot:
        snapshot['branches'] = {
            k: enrich_object(v, request) if v else None
            for k, v in snapshot['branches'].items()
        }
        for k, v in snapshot['branches'].items():
            if v and v['target_type'] == 'alias':
                branch = resolve_branch_alias(snapshot, v)
                if branch:
                    branch = enrich_object(branch, request)
                    v['target_url'] = branch['target_url']
    return snapshot


def enrich_origin(origin: Dict[str, Any],
                  request: Optional[HttpRequest] = None) -> Dict[str, Any]:
    """Enrich origin dict with link to its visits

    Args:
        origin: the origin as a dict
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched origin dict filled with an additional url
    """
    if 'url' in origin:
        origin['origin_visits_url'] = reverse(
            'api-1-origin-visits',
            url_args={'origin_url': origin['url']},
            request=request)

    return origin


def enrich_origin_visit(origin_visit: Dict[str, Any], *,
                        with_origin_link: bool, with_origin_visit_link: bool,
                        request: Optional[HttpRequest] = None
                        ) -> Dict[str, Any]:
    """Enrich origin visit dict with additional links

    Args:
        origin_visit: the origin visit as a dict
        with_origin_link: whether to add link to origin
        with_origin_visit_link: whether to add link to origin visit
        request: Absolute URIs will be generated if provided

    Returns:
        An enriched origin visit dict filled with additional urls
    """
    ov = origin_visit
    if with_origin_link:
        ov['origin_url'] = reverse('api-1-origin',
                                   url_args={'origin_url': ov['origin']},
                                   request=request)
    if with_origin_visit_link:
        ov['origin_visit_url'] = reverse('api-1-origin-visit',
                                         url_args={'origin_url': ov['origin'],
                                                   'visit_id': ov['visit']},
                                         request=request)
    snapshot = ov['snapshot']
    if snapshot:
        ov['snapshot_url'] = reverse('api-1-snapshot',
                                     url_args={'snapshot_id': snapshot},
                                     request=request)
    else:
        ov['snapshot_url'] = None
    return ov



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common import service
from swh.web.common.utils import reverse
from swh.web.config import get_config
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.utils import enrich_snapshot
from swh.web.api.views.utils import api_lookup


@api_route(r'/snapshot/(?P<snapshot_id>[0-9a-f]+)/', 'api-1-snapshot',
           checksum_args=['snapshot_id'])
@api_doc('/snapshot/')
@format_docstring()
def api_snapshot(request, snapshot_id):
    """
    .. http:get:: /api/1/snapshot/(snapshot_id)/

        Get information about a snapshot in the archive.

        A snapshot is a set of named branches, which are pointers to objects
        at any level of the Software Heritage DAG. It represents a full picture
        of an origin at a given time.

        As well as pointing to other objects in the Software Heritage DAG,
        branches can also be aliases, in which case their target is the name of
        another branch in the same snapshot, or dangling, in which case the
        target is unknown.

        A snapshot identifier is a salted sha1. See
        :func:`swh.model.identifiers.snapshot_identifier` in our data model
        module for details about how they are computed.

        :param sha1 snapshot_id: a snapshot identifier
        :query str branches_from: optional parameter used to skip branches
            whose name is lesser than it before returning them
        :query int branches_count: optional parameter used to restrain
            the amount of returned branches (default to 1000)
        :query str target_types: optional comma separated list parameter
            used to filter the target types of branch to return (possible
            values that can be contained in that list are ``content``,
            ``directory``, ``revision``, ``release``, ``snapshot`` or
            ``alias``)

        {common_headers}
        {resheader_link}

        :>json object branches: object containing all branches associated to
            the snapshot,for each of them the associated target type and id are
            given but also a link to get information about that target
        :>json string id: the unique identifier of the snapshot

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`,
        :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid snapshot identifier has been provided
        :statuscode 404: requested snapshot can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`snapshot/6a3a2cf0b2b90ce7ae1cf0a221ed68035b686f5a/`
    """

    snapshot_content_max_size = get_config()['snapshot_content_max_size']

    branches_from = request.GET.get('branches_from', '')
    branches_count = int(request.GET.get('branches_count',
                                         snapshot_content_max_size))
    target_types = request.GET.get('target_types', None)
    target_types = target_types.split(',') if target_types else None

    results = api_lookup(
        service.lookup_snapshot, snapshot_id, branches_from,
        branches_count, target_types,
        notfound_msg='Snapshot with id {} not found.'.format(snapshot_id),
        enrich_fn=enrich_snapshot,
        request=request)

    response = {'results': results, 'headers': {}}

    if results['next_branch'] is not None:
        response['headers']['link-next'] = reverse(
            'api-1-snapshot',
            url_args={'snapshot_id': snapshot_id},
            query_params={'branches_from': results['next_branch'],
                          'branches_count': branches_count,
                          'target_types': target_types},
            request=request)

    return response



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given

from swh.web.api.utils import enrich_revision
from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import reverse
from swh.web.tests.data import random_sha1
from swh.web.tests.strategies import revision


@given(revision())
def test_api_revision(api_client, archive_data, revision):
    url = reverse('api-1-revision', url_args={'sha1_git': revision})
    rv = api_client.get(url)

    expected_revision = archive_data.revision_get(revision)

    enrich_revision(expected_revision, rv.wsgi_request)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == expected_revision


def test_api_revision_not_found(api_client):
    unknown_revision_ = random_sha1()

    url = reverse('api-1-revision',
                  url_args={'sha1_git': unknown_revision_})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Revision with sha1_git %s not found.' % unknown_revision_
    }


@given(revision())
def test_api_revision_raw_ok(api_client, archive_data, revision):
    url = reverse('api-1-revision-raw-message',
                  url_args={'sha1_git': revision})
    rv = api_client.get(url)

    expected_message = archive_data.revision_get(revision)['message']

    assert rv.status_code == 200
    assert rv['Content-Type'] == 'application/octet-stream'
    assert rv.content == expected_message.encode()


def test_api_revision_raw_ko_no_rev(api_client):
    unknown_revision_ = random_sha1()

    url = reverse('api-1-revision-raw-message',
                  url_args={'sha1_git': unknown_revision_})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Revision with sha1_git %s not found.' % unknown_revision_
    }


@given(revision())
def test_api_revision_log(api_client, archive_data, revision):
    per_page = 10

    url = reverse('api-1-revision-log', url_args={'sha1_git': revision},
                  query_params={'per_page': per_page})

    rv = api_client.get(url)

    expected_log = archive_data.revision_log(revision, limit=per_page+1)
    expected_log = list(map(enrich_revision, expected_log,
                            [rv.wsgi_request] * len(expected_log)))

    has_next = len(expected_log) > per_page

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == (expected_log[:-1] if has_next else expected_log)

    if has_next:
        assert 'Link' in rv
        next_log_url = rv.wsgi_request.build_absolute_uri(
            reverse('api-1-revision-log',
                    url_args={'sha1_git': expected_log[-1]['id']},
                    query_params={'per_page': per_page}))
        assert next_log_url in rv['Link']


def test_api_revision_log_not_found(api_client):
    unknown_revision_ = random_sha1()

    url = reverse('api-1-revision-log',
                  url_args={'sha1_git': unknown_revision_})

    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Revision with sha1_git %s not found.' % unknown_revision_
    }
    assert not rv.has_header('Link')


@given(revision())
def test_api_revision_log_context(api_client, archive_data, revision):
    revisions = archive_data.revision_log(revision, limit=4)

    prev_rev = revisions[0]['id']
    rev = revisions[-1]['id']

    per_page = 10

    url = reverse('api-1-revision-log',
                  url_args={'sha1_git': rev,
                            'prev_sha1s': prev_rev},
                  query_params={'per_page': per_page})

    rv = api_client.get(url)

    expected_log = archive_data.revision_log(rev, limit=per_page)
    prev_revision = archive_data.revision_get(prev_rev)
    expected_log.insert(0, prev_revision)
    expected_log = list(map(enrich_revision, expected_log,
                            [rv.wsgi_request] * len(expected_log)))

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == expected_log


def test_api_revision_directory_ko_not_found(api_client, mocker):
    mock_rev_dir = mocker.patch(
        'swh.web.api.views.revision._revision_directory_by')
    mock_rev_dir.side_effect = NotFoundExc('Not found')

    rv = api_client.get('/api/1/revision/999/directory/some/path/to/dir/')

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Not found'
    }

    mock_rev_dir.assert_called_once_with(
        {'sha1_git': '999'},
        'some/path/to/dir',
        '/api/1/revision/999/directory/some/path/to/dir/',
        with_data=False
    )


def test_api_revision_directory_ok_returns_dir_entries(api_client, mocker):
    mock_rev_dir = mocker.patch(
        'swh.web.api.views.revision._revision_directory_by')
    stub_dir = {
        'type': 'dir',
        'revision': '999',
        'content': [
            {
                'sha1_git': '789',
                'type': 'file',
                'target': '101',
                'target_url': '/api/1/content/sha1_git:101/',
                'name': 'somefile',
                'file_url': '/api/1/revision/999/directory/some/path/'
                'somefile/'
            },
            {
                'sha1_git': '123',
                'type': 'dir',
                'target': '456',
                'target_url': '/api/1/directory/456/',
                'name': 'to-subdir',
                'dir_url': '/api/1/revision/999/directory/some/path/'
                'to-subdir/',
            }
        ]
    }

    mock_rev_dir.return_value = stub_dir

    rv = api_client.get('/api/1/revision/999/directory/some/path/')

    stub_dir['content'][0]['target_url'] = rv.wsgi_request.build_absolute_uri(
        stub_dir['content'][0]['target_url'])
    stub_dir['content'][0]['file_url'] = rv.wsgi_request.build_absolute_uri(
        stub_dir['content'][0]['file_url'])
    stub_dir['content'][1]['target_url'] = rv.wsgi_request.build_absolute_uri(
        stub_dir['content'][1]['target_url'])
    stub_dir['content'][1]['dir_url'] = rv.wsgi_request.build_absolute_uri(
        stub_dir['content'][1]['dir_url'])

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == stub_dir

    mock_rev_dir.assert_called_once_with(
        {'sha1_git': '999'},
        'some/path',
        '/api/1/revision/999/directory/some/path/',
        with_data=False
    )


def test_api_revision_directory_ok_returns_content(api_client, mocker):
    mock_rev_dir = mocker.patch(
        'swh.web.api.views.revision._revision_directory_by')
    stub_content = {
        'type': 'file',
        'revision': '999',
        'content': {
            'sha1_git': '789',
            'sha1': '101',
            'data_url': '/api/1/content/101/raw/',
        }
    }

    mock_rev_dir.return_value = stub_content

    url = '/api/1/revision/666/directory/some/other/path/'
    rv = api_client.get(url)

    stub_content['content']['data_url'] = rv.wsgi_request.build_absolute_uri(
        stub_content['content']['data_url'])

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == stub_content

    mock_rev_dir.assert_called_once_with(
        {'sha1_git': '666'}, 'some/other/path', url, with_data=False)


@given(revision())
def test_api_revision_uppercase(api_client, revision):
    url = reverse('api-1-revision-uppercase-checksum',
                  url_args={'sha1_git': revision.upper()})

    resp = api_client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('api-1-revision',
                           url_args={'sha1_git': revision})

    assert resp['location'] == redirect_url



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from corsheaders.middleware import ACCESS_CONTROL_ALLOW_ORIGIN
from hypothesis import given

from swh.model.identifiers import (
    persistent_identifier,
    CONTENT, DIRECTORY, ORIGIN, RELEASE, REVISION, SNAPSHOT
)
from swh.web.common import service
from swh.web.common.utils import reverse, resolve_swh_persistent_id
from swh.web.misc.badges import _badge_config, _get_logo_data
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.strategies import (
    content, directory, origin, release, revision, snapshot,
    unknown_content, unknown_directory, new_origin, unknown_release,
    unknown_revision, unknown_snapshot, invalid_sha1
)


@given(content())
def test_content_badge(client, content):
    _test_badge_endpoints(client, CONTENT, content['sha1_git'])


@given(directory())
def test_directory_badge(client, directory):
    _test_badge_endpoints(client, DIRECTORY, directory)


@given(origin())
def test_origin_badge(client, origin):
    _test_badge_endpoints(client, ORIGIN, origin['url'])


@given(release())
def test_release_badge(client, release):
    _test_badge_endpoints(client, RELEASE, release)


@given(revision())
def test_revision_badge(client, revision):
    _test_badge_endpoints(client, REVISION, revision)


@given(snapshot())
def test_snapshot_badge(client, snapshot):
    _test_badge_endpoints(client, SNAPSHOT, snapshot)


@given(unknown_content(), unknown_directory(), new_origin(),
       unknown_release(), unknown_revision(), unknown_snapshot(),
       invalid_sha1())
def test_badge_errors(client, unknown_content, unknown_directory, new_origin,
                      unknown_release, unknown_revision, unknown_snapshot,
                      invalid_sha1):
    for object_type, object_id in (
        (CONTENT, unknown_content['sha1_git']),
        (DIRECTORY, unknown_directory),
        (ORIGIN, new_origin['url']),
        (RELEASE, unknown_release),
        (REVISION, unknown_revision),
        (SNAPSHOT, unknown_snapshot)
    ):
        url_args = {
            'object_type': object_type,
            'object_id': object_id
        }
        url = reverse('swh-badge', url_args=url_args)
        resp = client.get(url)
        _check_generated_badge(resp, **url_args, error='not found')

        if object_type != ORIGIN:
            object_pid = persistent_identifier(object_type, object_id)
            url = reverse('swh-badge-pid',
                          url_args={'object_pid': object_pid})
            resp = client.get(url)
            _check_generated_badge(resp, **url_args, error='not found')

    for object_type, object_id in (
        (CONTENT, invalid_sha1),
        (DIRECTORY, invalid_sha1),
        (RELEASE, invalid_sha1),
        (REVISION, invalid_sha1),
        (SNAPSHOT, invalid_sha1)
    ):
        url_args = {
            'object_type': object_type,
            'object_id': object_id
        }
        url = reverse('swh-badge', url_args=url_args)
        resp = client.get(url)
        _check_generated_badge(resp, **url_args, error='invalid id')

        object_pid = f'swh:1:{object_type[:3]}:{object_id}'
        url = reverse('swh-badge-pid',
                      url_args={'object_pid': object_pid})
        resp = client.get(url)
        _check_generated_badge(resp, '', '', error='invalid id')


@given(origin(), release())
def test_badge_endpoints_have_cors_header(client, origin, release):
    url = reverse('swh-badge', url_args={'object_type': ORIGIN,
                                         'object_id': origin['url']})
    resp = client.get(url, HTTP_ORIGIN='https://example.org')
    assert resp.status_code == 200, resp.content
    assert ACCESS_CONTROL_ALLOW_ORIGIN in resp

    release_pid = persistent_identifier(RELEASE, release)
    url = reverse('swh-badge-pid', url_args={'object_pid': release_pid})
    resp = client.get(url, HTTP_ORIGIN='https://example.org')
    assert resp.status_code == 200, resp.content
    assert ACCESS_CONTROL_ALLOW_ORIGIN in resp


def _test_badge_endpoints(client, object_type, object_id):
    url_args = {'object_type': object_type,
                'object_id': object_id}
    url = reverse('swh-badge', url_args=url_args)
    resp = client.get(url)
    _check_generated_badge(resp, **url_args)
    if object_type != ORIGIN:
        pid = persistent_identifier(object_type, object_id)
        url = reverse('swh-badge-pid', url_args={'object_pid': pid})
        resp = client.get(url)
        _check_generated_badge(resp, **url_args)


def _check_generated_badge(response, object_type, object_id, error=None):
    assert response.status_code == 200, response.content
    assert response['Content-Type'] == 'image/svg+xml'

    if not object_type:
        object_type = 'object'

    if object_type == ORIGIN and error is None:
        link = reverse('browse-origin', url_args={'origin_url': object_id})
        text = 'repository'
    elif error is None:
        text = persistent_identifier(object_type, object_id)
        link = resolve_swh_persistent_id(text)['browse_url']
        if object_type == RELEASE:
            release = service.lookup_release(object_id)
            text = release['name']
    elif error == 'invalid id':
        text = 'error'
        link = f'invalid {object_type} id'
        object_type = 'error'
    elif error == 'not found':
        text = 'error'
        link = f'{object_type} not found'
        object_type = 'error'

    assert_contains(response, '<svg ')
    assert_contains(response, '</svg>')
    assert_contains(response, _get_logo_data())
    assert_contains(response, _badge_config[object_type]['color'])
    assert_contains(response, _badge_config[object_type]['title'])
    assert_contains(response, text)
    assert_contains(response, link)
